# -*- coding: utf-8 -*-
"""Manage the data of a field in a table."""
from pineboolib.core.utils import logging, utils_base
from pineboolib import interfaces

from typing import List, Optional, Union, Any, TYPE_CHECKING

from pineboolib.application.metadata import pnrelationmetadata

if TYPE_CHECKING:
    from pineboolib.application.metadata import pntablemetadata  # pragma: no cover

LOGGER = logging.get_logger(__name__)


class PNFieldMetaData(interfaces.IFieldMetaData):
    """PNFieldMetaData Class."""

    Serial = "serial"
    Unlock = "unlock"
    Check = "check"
    count_ = 0

    def __init__(self, *args, **kwargs) -> None:
        """
        Initialize the metadata with the collected data.

        @param n Field Name.
        @param to Alias ​​del campo, used in form labels.
        @param aN TRUE if it allows nulls (NULL), FALSE if it allows them (NOT NULL).
        @param _isPrimaryKey TRUE if it is a primary key, FALSE if it is not a primary key, be
                primary key implies being Index and Unique.
        @param t Field type.
        @param l Length of the field in characters, provided it is of type string
               of characters.
        @param c Indicates if the field is calculated.
        @param v Indicates if the field is visible.
        @param ed Indicates if the field is editable.
        @param pI Indicates the number of digits of the whole part.
        @param pD Indicates the number of decimals.
        @param iNX TRUE if the field is index.
        @param uNI TRUE if the field determines unique records.
        @param coun Indicates if it is an accountant. For automatic references.
        @param defValue Default value for the field.
        @param oT Indicates if the changes in the field are out of transaction.
        @param rX Regular expression used as a validation mask.
        @param vG Indicates if the field is visible in the grid of the table.
        @param gen Indicates if the field is generated.
        @param iCK Indicates if it is a composite key.
        """

        if len(args) == 1:
            self.inicializeFLFieldMetaData(args[0])
        else:
            self.inicializeNewFLFieldMetaData(*args, **kwargs)
        ++self.count_

    def inicializeFLFieldMetaData(self, other: "PNFieldMetaData") -> None:
        """Initialize by copying information from another metadata."""

        self.private = PNFieldMetaDataPrivate()
        self.copy(other)

    def inicializeNewFLFieldMetaData(
        self,
        field_name: str,
        alias: str,
        allow_null: bool,
        is_pk: bool,
        field_type: str,
        length: int = 0,
        is_calculated: bool = False,
        is_visible: bool = True,
        is_editable: bool = True,
        part_integer: int = 4,
        part_decimal: int = 0,
        is_index: bool = False,
        is_unique: bool = False,
        is_counter: bool = False,
        default_value: Optional[str] = None,
        out_transaction: bool = False,
        regular_expr: Optional[str] = None,
        is_visible_grid: bool = True,
        is_generated: bool = True,
        is_ck: bool = False,
    ) -> None:
        """
        Initialize with the information collected.

        @param field_name Field Name.
        @param alias Alias ​​del campo, used in form labels.
        @param allow_null TRUE if it allows nulls (NULL), FALSE if it allows them (NOT NULL).
        @param is_pk TRUE if it is a primary key, FALSE if it is not a primary key, be
                primary key implies being Index and Unique.
        @param field_type Field type.
        @param length Length of the field in characters, provided it is of type string
               of characters.
        @param is_calculated Indicates if the field is calculated.
        @param is_visible Indicates if the field is visible.
        @param is_editable Indicates if the field is editable.
        @param part_integer Indicates the number of digits of the whole part.
        @param part_decimal Indicates the number of decimals.
        @param is_index TRUE if the field is index.
        @param is_unique TRUE if the field determines unique records.
        @param is_counter Indicates if it is an accountant. For automatic references.
        @param default_value Default value for the field.
        @param out_transaction Indicates if the changes in the field are out of transaction.
        @param regular_expr Regular expression used as a validation mask.
        @param is_visible_grid Indicates if the field is visible in the grid of the table.
        @param is_generated Indicates if the field is generated.
        @param is_ck Indicates if it is a composite key.
        """
        self.private = PNFieldMetaDataPrivate(
            field_name,
            alias,
            allow_null,
            is_pk,
            field_type,
            length,
            is_calculated,
            is_visible,
            is_editable,
            part_integer,
            part_decimal,
            is_index,
            is_unique,
            is_counter,
            default_value,
            out_transaction,
            regular_expr,
            is_visible_grid,
            is_generated,
            is_ck,
        )

    def name(self) -> str:
        """
        Get the name of the field.

        @return Field Name.
        """
        return self.private._field_name

    def setName(self, name: str) -> None:
        """
        Set the name for the field.

        @param n Field Name
        """
        self.private._field_name = name

    def alias(self) -> str:
        """
        Get the alias of the field.

        @return Alias Name.
        """

        return utils_base.AQTT(self.private.alias_)

    def allowNull(self) -> bool:
        """
        Get if it allows nulls.

        @return TRUE if it allows nulls, FALSE otherwise
        """

        return self.private._allow_null

    def isPrimaryKey(self) -> bool:
        """
        Get if it is primary key.

        @return TRUE if it is primary key, FALSE otherwise
        """

        return self.private.is_primary_key

    def setIsPrimaryKey(self, value: bool) -> None:
        """
        Set if it is primary key.

        @return TRUE if it is primary key, FALSE otherwise
        """

        self.private.is_primary_key = value

    def isCompoundKey(self) -> bool:
        """
        Get if it is a composite key.

        @return TRUE if it is a composite key, FALSE otherwise
        """

        return self.private.is_compound_key or False

    def type(self) -> str:
        """
        Return the type of the field.

        @return The field type.
        """
        return self.private.type_

    def length(self) -> int:
        """
        Get the length of the field.

        @return The length of the field.
        """

        return int(self.private.length_ or 0)

    def calculated(self) -> bool:
        """
        Get if the field is calculated.

        @return TRUE if the field is calculated, FALSE otherwise
        """

        return self.private.calculated_

    def setCalculated(self, calculated: bool) -> None:
        """
        Set if the field is calculated.

        @param calculated Value TRUE if you want to set the field as calculated, FALSE otherwise.
        """
        self.private.calculated_ = calculated

    def editable(self) -> bool:
        """
        Get if the field is editable.

        @return TRUE if the field is editable, FALSE otherwise
        """
        return self.private.editable_ or False

    def setEditable(self, editable: bool) -> None:
        """
        Set whether the field is editable.

        @param ed Value TRUE if you want the field to be editable, FALSE otherwise.
        """
        self.private.editable_ = editable

    def visible(self) -> bool:
        """
        Get if the field is visible.

        @return TRUE if the field is visible, FALSE otherwise
        """
        return self.private.visible_

    def visibleGrid(self) -> bool:
        """
        Get if the field is visible in the grid of the table.

        @return TRUE if the field is visible in the grid of the table, FALSE otherwise
        """
        return self.private._visible_grid

    def generated(self) -> bool:
        """@return TRUE if the field is generated, that is, it is included in the queries."""

        return self.private.generated_

    def setGenerated(self, generated: bool) -> None:
        """Set a field as generated."""

        self.private.generated_ = generated

    def setVisible(self, visible: bool) -> None:
        """
        Set if the field is visible.

        @param v Value TRUE if you want to make the field visible, FALSE otherwise.
        """

        self.private.visible_ = visible

    def setVisibleGrid(self, visible_grid: bool) -> None:
        """
        Set whether the field is visible in the grid of the table.

        @param v Value TRUE if you want to make the field visible, FALSE otherwise.
        """

        self.private._visible_grid = visible_grid

    def partInteger(self) -> int:
        """
        Get the number of digits of the whole part.

        @return The number of digits of the entire part of the field.
        """

        return int(self.private._part_integer or 0)

    def partDecimal(self) -> int:
        """
        Get the number of digits of the decimal part.

        @return The number of digits of the decimal part of the field.
        """

        return int(self.private._part_decimal or 0)

    def isCounter(self) -> bool:
        """
        Get if the field is a counter.

        @return TRUE if the field is a reference with counter
        """
        return self.private.contador_

    def isIndex(self) -> bool:
        """
        Get if the field is index.

        @return TRUE if the field is index, FALSE otherwise
        """
        return self.private._is_index

    def isUnique(self) -> bool:
        """
        Get if the field determines unique records.

        @return TRUE if the field determines unique records, FALSE otherwise
        """
        return self.private._is_unique

    def addRelationMD(self, relation: "pnrelationmetadata.PNRelationMetaData") -> None:
        """
        Add a relationship with another table for this field.

        Add a new FLRelationMetaData object to the list of relationships for this field.
        Note that for one field there can only be one single ratio of type M1 (many to one), so in
        in case you want to add several relationships of this type for the field only the first one will be taken into account.
        Type 1M relationships (one to many) may all exist those necessary. See FLRelationMetaData :: Cardinality.

        @param r FlRelationMetaData object with the definition of the relationship to add.
        """

        if not self.private._field_name:
            LOGGER.warning("addRelationMD: no fieldName")
            return

        if relation.cardinality() == pnrelationmetadata.PNRelationMetaData.RELATION_M1:
            if self.private._relation_m1:
                LOGGER.warning(
                    "addRelationMD: Se ha intentado crear más de una relación muchos a uno para el mismo campo"
                )
                return
            self.private._relation_m1 = relation
        else:
            self.private._relation_list.append(relation)

        relation.setField(self.private._field_name)

    def relationList(self) -> List["pnrelationmetadata.PNRelationMetaData"]:
        """
        To get the list of relationship definitions.

        Does not include the M1 relationship

        @return Object with the list of deficits in the field relations
        """

        return self.private._relation_list

    def relationM1(self) -> Optional["pnrelationmetadata.PNRelationMetaData"]:
        """
        Get the many-to-one relationship for this field.

        Not included in relationList ()

        @return Object FLRelationMetaData with the description of the relationship many to one for this field.
        """
        return self.private._relation_m1

    def setAssociatedField(
        self, r_or_name: Union[str, "interfaces.IFieldMetaData"], field_name: str
    ) -> None:
        """
        Set an associated field for this field, and the name of the foreign table field to use to filter.

            according to the value of the associated field.

        @param r FLFieldMetaData object or Name that defines the field to be associated with this
        @param f Name of the field to apply the filter

        """
        name = r_or_name if isinstance(r_or_name, str) else r_or_name.name()

        self.private.associated_field_name = name
        self.private.associated_field_filter_to = field_name

    def associatedField(self) -> Optional["PNFieldMetaData"]:
        """
        Return the associated field for this field.

        @return FLFieldMetaData object that defines the field associated with it, or 0 if there is no associated field.
        """
        table_mtd = self.metadata()
        return (
            table_mtd.field(self.private.associated_field_name) if table_mtd is not None else None
        )

    def associatedFieldFilterTo(self) -> str:
        """
        Return the name of the field to be filtered according to the associated field.

        @return Field name of the foreign table M-1, to which the filter must be applied according to the value of the associated field.
        """
        return self.private.associated_field_filter_to

    def associatedFieldName(self) -> Optional[str]:
        """
        Return the name of the associated field this.

        @return Name of the associated field.
        """

        return self.private.associated_field_name

    def defaultValue(self) -> Optional[Any]:
        """
        Return the default value for the field.

        @return Value that is assigned to the field by default
        """

        if self.private._default_value == "null":
            self.private._default_value = None

        elif self.private.type_ in ("bool", "unlock") and isinstance(
            self.private._default_value, str
        ):
            self.private._default_value = self.private._default_value in ["true", "True"]

        return self.private._default_value

    def outTransaction(self) -> bool:
        """
        Return if the field is modified out of transaction.

        @return TRUE if the field is modified out of transaction, FALSE otherwise
        """
        return self.private._out_transaction or False

    def regExpValidator(self) -> Optional[str]:
        """
        Return the regular expression that serves as a validation mask for the field.

        @return Character string containing a regular expression, used as
            mask to validate the values ​​entered in the field
        """
        return self.private._reg_exp_validator

    def optionsList(self) -> List[str]:
        """
        Return the list of options for the field.

        @return List of field options
        """
        return self.private._options_list

    def getIndexOptionsList(self, name: str) -> Optional[int]:
        """
        Return the index of a given field.

        @return List of field options.
        """
        return (
            self.private._options_list.index(name) if name in self.private._options_list else None
        )

    def setOptionsList(self, options_list: str) -> None:
        """
        Set the list of options for the field.

        @param ol Text string with options for the field.
        """

        self.private._options_list = [
            utils_base.AQTT(componente).strip()
            for componente in options_list.split(
                ";" if options_list.find("QT_TRANSLATE") > -1 else ","
            )
        ]

    def isCheck(self) -> bool:
        """
        Get if the field is of type Check.
        """

        return self.private.type_ == self.Check

    def hasOptionsList(self) -> bool:
        """
        Get if the field has a list of options.
        """

        return True if self.private._options_list else False

    def fullyCalculated(self) -> bool:
        """
        Return if a field is fully calculated.
        """

        return self.private._fully_calculated

    def setFullyCalculated(self, calculated: bool) -> None:
        """
        Specify if a field is fully calculated.
        """

        self.private._fully_calculated = calculated
        if calculated:
            self.private.generated_ = True

    def trimmed(self) -> bool:
        """Return if a field is trimmed."""

        return self.private.trimmed_

    def setTrimmed(self, trimmed: bool) -> None:
        """Specify if a field is trimmed."""

        self.private.trimmed_ = trimmed

    def setMetadata(self, mtd: "pntablemetadata.PNTableMetaData") -> None:
        """
        Set the PNTableMetaData object to which it belongs.
        """
        self.private.mtd_ = mtd

    def metadata(self) -> Optional["pntablemetadata.PNTableMetaData"]:
        """
        Get the FLTableMetaData object to which it belongs.
        """

        return self.private.mtd_

    def flDecodeType(self, fltype_=None) -> Optional[str]:
        """
        Get the type of the field converted to an equivalent type of the QVariant class.
        """

        _type = None
        # print("Decode", fltype)

        if fltype_ == "int":
            _type = "int"
        elif fltype_ in ("serial", "uint"):
            _type = "uint"
        elif fltype_ in ("bool", "unlock"):
            _type = "bool"
        elif fltype_ == "double":
            _type = "double"
        elif fltype_ == "time":
            _type = "time"
        elif fltype_ == "date":
            _type = "date"
        elif fltype_ in ("string", "pixmap", "stringlist", "timestamp"):
            _type = "string"
        elif fltype_ == "bytearray":
            _type = "bytearray"
        elif fltype_ == "json":
            _type = "json"

        # print("Return", _type)
        return _type

    def searchOptions(self) -> List[str]:
        """
        Return different search options for this field.

        @return list of different options
        """

        return self.private._search_options

    def setSearchOptions(self, options_list: str) -> None:
        """
        Set the list of options for the field.

        @param ol Text string with options for the field.
        """

        self.private._search_options = options_list.split(",")

    def copy(self, other: "PNFieldMetaData") -> None:
        """
        Copy the metadata of another pnfieldmetadata.
        """

        if other is self:
            return

        other_private = other.private

        if other_private._relation_m1:
            self.private._relation_m1 = other_private._relation_m1

        self.private.clearRelationList()

        if other_private._relation_list:
            for relation in other_private._relation_list:
                self.private._relation_list.append(relation)

        self.private._field_name = other_private._field_name
        self.private.alias_ = other_private.alias_
        self.private._allow_null = other_private._allow_null
        self.private.is_primary_key = other_private.is_primary_key
        self.private.type_ = other_private.type_
        self.private.length_ = other_private.length_
        self.private.calculated_ = other_private.calculated_
        self.private._fully_calculated = other_private._fully_calculated
        self.private.trimmed_ = other_private.trimmed_
        self.private.visible_ = other_private.visible_
        self.private.editable_ = other_private.editable_
        self.private._part_decimal = other_private._part_decimal
        self.private._part_integer = other_private._part_integer
        self.private._is_index = other_private._is_index
        self.private._is_unique = other_private._is_unique
        self.private.contador_ = other_private.contador_
        self.private.associated_field_name = other_private.associated_field_name
        self.private.associated_field_filter_to = other_private.associated_field_filter_to
        self.private._default_value = other_private._default_value
        self.private._options_list = other_private._options_list
        self.private._out_transaction = other_private._out_transaction
        self.private._reg_exp_validator = other_private._reg_exp_validator
        self.private._visible_grid = other_private._visible_grid
        self.private.generated_ = other_private.generated_
        self.private.is_compound_key = other_private.is_compound_key

        # self = copy.deepcopy(other)

    def formatAssignValue(self, field_name: str, value: Any, upper: bool) -> str:
        """
        Return the correct comparison for a given field.
        """

        if value is None or not field_name:
            return "1 = 1"

        is_text = self.type() in ("string", "time", "date", "pixmap", "timestamp")

        format_value: Any = "'%s'" % value if is_text else value

        if upper and is_text:
            field_name = "upper(%s)" % field_name
            format_value = format_value.upper()

        return "%s = %s" % (field_name, format_value)

    def __len__(self) -> int:
        """Return the length of a field."""

        return int(self.private.length_ or 0)

    def __bool__(self) -> bool:
        """Return if exists."""

        return True


class PNFieldMetaDataPrivate(object):
    """PNFieldMetaDataPrivate Class."""

    """
    Nombre del campo en la tabla
    """

    _field_name: str

    """
    Alias o mote para el campo, usado como
    etiqueta de campos en los formularios
    """
    alias_: str

    """
    Almacena si el campo permite ser nulo
    """
    _allow_null: bool

    """
    Almacena si el campo es clave primaria
    """
    is_primary_key: bool

    """
    Tipo del campo
    """
    type_: str

    """
    Longitud del campo
    """
    length_: int

    """
    Indica si el campo es calculado de forma diferida.
    Esto indica que el campo se calcula al editar o insertar un registro, en el commit.
    """
    calculated_: bool

    """
    Indica si el campo es totalmente calculado.
    Esto indica que el valor campo del campo es dinámico y se calcula en cada refresco.
    Un campo totalmente calculado implica que es generado.
    """
    _fully_calculated: bool

    """
    Indica que al leer el campo de la base de datos los espacios mas a la derecha
    son eliminados.
    """
    trimmed_: bool

    """
    Indica si el campo es visible
    """
    visible_: bool

    """
    Indica si el campo es editable
    """
    editable_: bool

    """
    Indica el número de dígitos de la parte entera
    """
    _part_integer: int

    """
    Indica el númeor de dígitos de la parte decimal
    """
    _part_decimal: int

    """
    Indica si el campo es índice
    """
    _is_index: bool

    """
    Indica si el campo es único
    """
    _is_unique: bool

    """
    Indica si el campo es un contador de referencia y abanq en el
    momento de insertar un registro debe intentar calcular cual sería el
    siguiente numero.

    @author Andrés Otón Urbano (andresoton@eresmas.com)
    """
    contador_: bool

    """
    Lista de relaciones para este campo
    """
    _relation_list: List["pnrelationmetadata.PNRelationMetaData"]

    """
    Mantiene, si procede, la relación M1 (muchos a uno)
    para el campo (solo puede haber una relacion de este tipo para un campo)
    """
    _relation_m1: Optional["pnrelationmetadata.PNRelationMetaData"]

    """
    Asocia este campo con otro, para efectuar filtros en búsquedas.

    El campo que se asocia a este debe tener una relación M-1.
    Este campo también debe tener una relación M-1. Al asociar un campo a este,
    las búsquedas mediante los botones de búsqueda en los formularios de edición
    de registros vendrán condicionadas por el valor del campo asociado en el
    momento de realizar dicha búsqueda. Cuando se realiza una búsqueda para
    este campo la tabla relacionada con él (M-1) será mostrada para elegir un
    registro de todos los posibles, en el caso normal se muestran todos los registros,
    pero cuando se asocia un campo sólo se muestran aquellos registros que satisfagan el
    valor del campo asociado. Ejemplo : En la tabla albaranes asociamos el campo
    'codemporig' al campo 'codalmorig' (NO es lo mismo que asociar 'codalmorig'
    a 'codemporig') cuando abrimos el formulario de albaranes elegimos una empresa
    origen (codemporig), cuando vayamos a elegir un almacen origen (codalmorig) sólo
    se podrá elegir entre los almacenes que son de la empresa origen , ya que el formulario
    de búsqueda sólo se mostrarán los almacenes cuyo código de empresa
    (ver FLFieldMetaData::associated_field_filter_to) sea igual al valor de la empresa origen
    elegida (codemporig)
    """
    associated_field_name: str

    """
    Nombre del campo que se debe filtra según el campo asociado.

    Esta propiedad sólo tiene sentido cuando hay un campo asociado a este,
    ver FLFieldMetaData ::associatedField_ , y si ese campo tiene una relacion M-1. Indica
    el nombre del campo de la tabla foránea en la relación M-1, que se debe utilizar para filtrar
    los registros según el valor del campo asociado. Ejemplo : En la tabla albaranes asociamos el campo
    'codemporig' al campo 'codalmorig' (NO es lo mismo que asociar 'codalmorig'
    a 'codemporig'), e indicamos que el campo de filtro es 'codempresa' de la tabla relacionada M-1 con el
    campo 'codalmorig' (Almacenes) . Cuando abrimos el formulario de albaranes elegimos una empresa
    origen (codemporig), cuando vayamos a elegir un almacen origen (codalmorig) sólo se podrá elegir
    entre los almacenes que son de la empresa origen, ya que el formulario de búsqueda sólo se mostrarán
    los almacenes cuyo código de empresa (el campo indicado de filtro ) sea igual al valor de la empresa
    origen elegida (codemporig)
    """
    associated_field_filter_to: str

    """
    Valor por defecto para el campo
    """
    _default_value: Any

    """
    Lista de opciones para el campo
    """
    _options_list: List[str]

    """
    Indica si las modificaciones del campo se hacen fuera de cualquier transaccion.

    Al estar activado este flag, todos los cambios en el valor de este campo se
    realizan fuera de la transaccion y de forma exclusiva. Es decir los cambios
    realizados en el campo son inmediatamente reflejados en la tabla sin esperar a
    que se termine transaccion, y de forma exclusiva (bloqueando el registro al que
    pertenece el campo mientras se modifica). Esto permite en el acto hacer visibles
    para todas las demas conexiones de la base de datos los cambios realizados en un campo.
    Hay que tener en cuenta que al tener el campo esta caracteristica especial de modificarse
    fuera de la transaccion, el "rollback" no tendra efecto sobre los cambios realizados
    en el y siempre permanecera en la base de datos la ultima modificacion efectuada en
    el campo.
    """
    _out_transaction: bool

    """
    Almacena la expresion regular que sirve como mascara de validacion para el campo.
    """
    _reg_exp_validator: Any
    _visible_grid: bool  # true
    generated_: bool
    is_compound_key: bool
    _search_options: List[str]
    mtd_: Optional["pntablemetadata.PNTableMetaData"]

    def __init__(self, *args, **kwargs) -> None:
        """Initialize the class."""

        self._field_name = ""
        self.alias_ = ""
        self._allow_null = False
        self.is_primary_key = False
        self.type_ = ""
        self.length_ = 0
        self.calculated_ = False
        self._fully_calculated = False
        self.trimmed_ = False
        self.visible_ = False
        self.editable_ = False
        self._part_integer = 0
        self._part_decimal = 0
        self._is_index = False
        self._is_unique = False
        self.contador_ = False
        self._relation_list = []
        self._relation_m1 = None
        self.associated_field_name = ""
        self.associated_field_filter_to = ""
        self._default_value = None
        self._options_list = []
        self._out_transaction = False
        self._reg_exp_validator = ""
        self._visible_grid = True
        self.generated_ = False
        self.is_compound_key = False
        self._search_options = []
        self.mtd_ = None

        if not args:
            self.inicializeEmpty()
        else:
            self.inicialize(*args, **kwargs)

    def inicializeEmpty(self) -> None:
        """Fill in the data without information."""

        self._relation_list = []
        self._relation_m1 = None
        self.associated_field_filter_to = ""
        self.associated_field_name = ""
        self.mtd_ = None

    def inicialize(
        self,
        name: str,
        alias: str,
        allow_null: bool,
        is_primary_key: bool,
        type_: str,
        length_: int,
        calculated: bool,
        visible: bool,
        editable: bool,
        part_integer: int,
        part_decimal: int,
        is_index: bool,
        is_unique: bool,
        coun: bool,
        default_value: Optional[str],
        out_transaction: bool,
        regular_expression: Optional[str],
        visible_grid: bool,
        gen: bool,
        is_compound_key: bool,
    ) -> None:
        """
        Fill in the data with information.

        @param name Field Name.
        @param alias ​​used in form labels.
        @param allow_null TRUE if it allows nulls (NULL), FALSE if it allows them (NOT NULL).
        @param is_primary_key TRUE if it is a primary key, FALSE if it is not a primary key, be
                primary key implies being Index and Unique.
        @param type_ Field type.
        @param length_ Length of the field in characters, provided it is of type string
               of characters.
        @param calculated Indicates if the field is calculated.
        @param visible Indicates if the field is visible.
        @param ed Indicates if the field is editable.
        @param part_integer Indicates the number of digits of the whole part.
        @param part_decimal Indicates the number of decimals.
        @param is_index TRUE if the field is index.
        @param is_unique TRUE if the field determines unique records.
        @param coun Indicates if it is an accountant. For automatic references.
        @param default_value Default value for the field.
        @param out_transaction Indicates if the changes in the field are out of transaction.
        @param regular_expression Regular expression used as a validation mask.
        @param visible_grid Indicates if the field is visible in the grid of the table.
        @param gen Indicates if the field is generated.
        @param is_compound_key Indicates if it is a compound key.
        """

        self._field_name = name.lower()
        self.alias_ = alias
        if calculated:
            self._allow_null = True
        else:
            self._allow_null = allow_null
        self.is_primary_key = is_primary_key
        self.type_ = type_
        self.length_ = length_
        self.calculated_ = calculated
        self.visible_ = visible
        self.editable_ = editable
        self._part_integer = part_integer
        self._part_decimal = part_decimal
        self._is_index = is_index
        self._is_unique = is_unique
        self.contador_ = coun
        self._relation_list = []
        self._relation_m1 = None
        self.associated_field_filter_to = ""
        self.associated_field_name = ""
        self._default_value = default_value
        self._out_transaction = out_transaction
        self._reg_exp_validator = regular_expression
        self._visible_grid = visible_grid
        self.generated_ = gen
        self.is_compound_key = is_compound_key
        self.mtd_ = None
        self._fully_calculated = False
        self.trimmed_ = False
        self._options_list = []

        if not self.type_:
            if self._part_decimal > 0:
                self.type_ = "double"
            elif self.length_ > 0:
                self.type_ = "string"
            else:
                self.type_ = "uint"
            LOGGER.info(
                "%s:: El campo %s no tiene especificado tipo y se especifica tipo %s",
                __name__,
                self._field_name,
                self.type_,
            )

        if self.type_ != "string" or self.length_ < 0:
            self.length_ = 0
        if self.type_ != "double" or self._part_integer < 0:
            self._part_integer = 0
        if self.type_ != "double" or self._part_decimal < 0:
            self._part_decimal = 0

    def __del_(self):
        """
        Delete properties when deleted.
        """
        self.clearRelationList()
        self._relation_m1 = None

    def clearRelationList(self) -> None:
        """
        Clear the list of relationship definitions.
        """
        self._relation_list.clear()



"""
Tests for known bugs on qsa.
"""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.application import types
from pineboolib.application.parsers.parser_qsa.postparse import pythonify_string as qs2py
from pineboolib.qsa import qsa


class TestKnownBugs(unittest.TestCase):
    """Test known bugs."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_isloadedmodule(self) -> None:
        """Test bug where sys.isLoadedModule seems to be None."""

        fn_test = types.function("module", "return sys.isLoadedModule(module);")
        self.assertEqual(fn_test("random_module"), False)
        self.assertEqual(fn_test("sys"), True)

    def test_mid(self) -> None:
        """Test str.mid(5, 2)."""

        value_1 = 'var cadena:String = "abcdefg";\ncadena.mid(5, 2);'
        self.assertEqual(qs2py(value_1), 'cadena: str = "abcdefg"\ncadena[5 : 5 + 2]\n')

        value_2 = 'var cadena:String = "abcdefg";\ncadena.mid(5);'
        self.assertEqual(qs2py(value_2), 'cadena: str = "abcdefg"\ncadena[0 + 5 :]\n')

    def test_system_get_env(self) -> None:
        """Check that qsa.System.getenv () works correctly."""

        qsa.System.setenv("TEST_PINEBOO", "¡hola!")

        self.assertEqual(qsa.System.getenv("TEST_PINEBOO"), "¡hola!")

    def test_self(self) -> None:
        """Test self."""

        value = "aqApp.self()"
        self.assertEqual(qs2py(value), "qsa.aqApp.self_()\n")

    def test_index_of(self) -> None:
        """Check indexOf translation."""

        value = """var text_:String = "test@test.test";\nif (text_.indexOf("@") == -1){\ndebug("ok");}"""
        self.assertEqual(
            qs2py(value),
            """text_: str = "test@test.test"\nif text_.index("@") == -1:\n    qsa.debug("ok")\n""",
        )

    def test_reg_exp(self) -> None:
        """Test regExp parser."""
        value = """var reg_exp:RegExp = new RegExp( "''" );\nreg_exp.global = true;"""
        self.assertEqual(
            qs2py(value), """reg_exp: "qsa.RegExp" = qsa.RegExp("''")\nreg_exp.global_ = True\n"""
        )

        value2 = """var reg_exp:RegExp = new RegExp( " " );\nreg_exp.global = true;\nvar texto:String = "UNO DOS".replace(reg_exp, "_");"""
        self.assertEqual(
            qs2py(value2),
            """reg_exp: \"qsa.RegExp\" = qsa.RegExp(" ")\nreg_exp.global_ = True\ntexto: str = qsa.replace("UNO DOS", reg_exp, "_")\n""",
        )

        value3 = """var reg_exp:RegExp = new RegExp( " " );\nreg_exp.global = true;\n
        var texto:String = "UNO DOS".replace(reg_exp, "_").lower();"""
        self.assertEqual(
            qs2py(value3),
            """reg_exp: \"qsa.RegExp\" = qsa.RegExp(" ")\nreg_exp.global_ = True\n"""
            + """texto: str = qsa.replace("UNO DOS", reg_exp, "_").lower()\n""",
        )

    def test_from_project(self) -> None:
        """Test from_project parser."""
        value_1 = (
            'qsa.sys.AQTimer.singleShot(0, qsa.from_project("formescarrosproveedor").reject)\n'
        )
        self.assertEqual(qs2py("sys.AQTimer.singleShot(0, formescarrosproveedor.reject);"), value_1)
        value_2 = (
            'qsa.sys.AQTimer.singleShot(0, qsa.from_project("formRecordflmodules").iface.prueba)\n'
        )
        self.assertEqual(
            qs2py("sys.AQTimer.singleShot(0, formRecordflmodules.iface.prueba);"), value_2
        )
        value_3 = 'qsa.sys.AQTimer.singleShot(0, qsa.from_project("flfactppal").iface.test)\n'
        self.assertEqual(qs2py("sys.AQTimer.singleShot(0, flfactppal.iface.test);"), value_3)

        value_4 = '_fP: Any = qsa.from_project("flfactppal").iface\n'
        self.assertEqual(qs2py("var _fP = flfactppal.iface;"), value_4)

    def test_multiples_if(self) -> None:
        """Check stackoverflow when parsing."""

        cadena = """var fN : String ="uno0";
                    if (fN == "uno0") {
                        debug(fN);
                        }
                    else if (fN == "dos0") {
                        debug(fN)
                        }
                    else if (fN == "tres0") {
                        debug(fN);
                        }
                    else if (fN == "cuatro0") {
                        debug(fN);
                        }
                    else if (fN == "cinco0") {
                        debug(fN);
                        }
                    else if (fN == "seis0") {
                        debug(fN);
                        }
                    else if (fN == "siete0") {
                        debug(fN);
                        }
                    else if (fN == "ocho0") {
                            debug(fN);
                            }
                    else if (fN == "nueve0") {
                        debug(fN);
                        }
                    else if (fN == "diez0") {
                        debug(fN);
                        }
                    else if (fN == "uno0") {
                        debug(fN);
                        }
                    else if (fN == "dos0") {
                        debug(fN);
                        }
                    else if (fN == "tres0") {
                        debug(fN);
                        }
                    else if (fN == "cuatro0") {
                        debug(fN);
                        }
                    else if (fN == "cinco0") {
                        debug(fN);
                        }
                    else if (fN == "seis0") {
                        debug(fN);
                        }
                    else if (fN == "siete0") {
                        debug(fN);
                        }
                    else if (fN == "ocho0") {
                            debug(fN);
                            }
                    else if (fN == "nueve0") {
                        debug(fN);
                        }
                    else if (fN == "diez0") {
                        debug(fN);
                        }
                    else if (fN == "dos0") {
                        debug(fN);
                        }
                    else if (fN == "tres0") {
                        debug(fN);
                        }
                    else if (fN == "cuatro0") {
                        debug(fN);
                        }
                    else if (fN == "cinco0") {
                        debug(fN);
                        }
                    else if (fN == "seis0") {
                        debug(fN);
                        }
                    else if (fN == "siete0") {
                        debug(fN);
                        }
                    else if (fN == "ocho0") {
                            debug(fN);
                            }
                    else if (fN == "nueve0") {
                        debug(fN);
                        }
                    else if (fN == "diez0") {
                        debug(fN);
                        }
                    else if (fN == "uno0") {
                        debug(fN);
                        }
                    else if (fN == "dos0") {
                        debug(fN);
                        }
                    else if (fN == "tres0") {
                        debug(fN);
                        }
                    else if (fN == "cuatro0") {
                        debug(fN);
                        }
                    else if (fN == "cinco0") {
                        debug(fN);
                        }
                    else if (fN == "seis0") {
                        debug(fN);
                        }
                    else if (fN == "siete0") {
                        debug(fN);
                        }
                    else if (fN == "ocho0" || fN == "once0") {
                            debug(fN);
                            }
                    else if (fN == "nueve0") {
                        debug(fN);
                        }
                    else if (fN == "diez0") {
                        debug(fN);
                        }

                    return fN;"""

        cadena_result = qs2py(cadena)
        self.assertFalse(cadena_result.find("not-known-seq") > -1)

    def test_from_project_2(self) -> None:
        """Test from_project('sys')."""

        self.assertTrue(hasattr(qsa.from_project("sys"), "iface"))

    def test_flsqlquery_date_null(self) -> None:
        """Test FLSqlQuery date null."""

        cursor = qsa.FLSqlCursor("fltest")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        self.assertTrue(cursor.commitBuffer())

        qry = qsa.FLSqlQuery()
        qry.setTablesList("fltest")
        qry.setSelect("date_field")
        qry.setFrom("fltest")
        qry.setWhere("1=1 LIMIT 1")
        self.assertTrue(qry.exec_())
        self.assertTrue(qry.first())
        self.assertEqual(str(qry.value(0)), "")

    def test_args(self) -> None:
        """Test translate with args parse."""

        cadena = qs2py(
            """var qryRecargo: FLSqlQuery = new FLSqlQuery;
        var res = util.translate("scripts", "Uno %1 para %2. ¿Desea continuar?")
        .arg(qryRecargo.value("f.codigo"))
        .arg(qryRecargo.value("f.nombrecliente"))"""
        )

        cadena_result = """qryRecargo: \"qsa.FLSqlQuery\" = qsa.FLSqlQuery()
res: Any = qsa.util.translate("scripts", "Uno %s para %s. ¿Desea continuar?") % (
    str(qryRecargo.value("f.codigo")),
    str(qryRecargo.value("f.nombrecliente")),\n)\n"""

        self.assertEqual(cadena, cadena_result)

    def test_args_str_int(self) -> None:
        """Test argStr and argInt."""

        one = """var res= util.translate("scripts", "Hola %1").argStr("Uno");"""
        result_one = """res: Any = qsa.util.translate("scripts", "Hola %s") % (str("Uno"))\n"""

        self.assertEqual(qs2py(one), result_one)

        two = """var res= util.translate("scripts", "Hola %1").argInt("Dos");"""
        result_two = """res: Any = qsa.util.translate("scripts", "Hola %s") % (str("Dos"))\n"""

        self.assertEqual(qs2py(two), result_two)

    def test_array_functions(self) -> None:
        """Test functions array conversion."""

        one = """var cacheFunsEdi = [];
        cacheFunsEdi["uno"]("dos");"""

        result_one = """cacheFunsEdi: Any = qsa.Array()\ncacheFunsEdi["uno"]("dos")\n"""
        self.assertEqual(qs2py(one), result_one)

    def test_number_min_value_and_max_value(self) -> None:
        """Test number min and max value."""
        qsa_one = "var min : Number = Number.MIN_VALUE;"
        result_one = "min = qsa.Number.MIN_VALUE\n"
        self.assertEqual(qs2py(qsa_one), result_one)

        qsa_two = "var max : Number = Number.MAX_VALUE;"
        result_two = "max = qsa.Number.MAX_VALUE\n"
        self.assertEqual(qs2py(qsa_two), result_two)

        self.assertTrue(qsa.Number.MIN_VALUE < 0)
        self.assertTrue(qsa.Number.MAX_VALUE > 0)

    def test_attribute_dict(self) -> None:
        """Test attribute dict converison."""

        att1 = qsa.AttributeDict(
            {"key_prueba1": ("valor_prueba1"), "key_prueba2": ("valor_prueba2")}
        )
        self.assertEqual(att1, {"key_prueba1": "valor_prueba1", "key_prueba2": "valor_prueba2"})

    def test_array_concat(self) -> None:
        """Test array concat simple."""
        array = qsa.Array().concat({"uno": "uno"}, {"dos": "dos"})
        self.assertEqual(array, {"uno": "uno", "dos": "dos"})
        array2 = qsa.Array().concat(["1", "2", "3"], ["5", "6"])
        self.assertEqual(array2, ["1", "2", "3", "5", "6"])

    def test_array_concat2(self) -> None:
        """Test array concat extended."""
        self.maxDiff = None
        codigo = """function getRules()
                    {
                        return new Array().concat(
                            this.getCrudRules('pedidoscli', 'Pedidos de cliente'),
                            [
                                {
                                    "idregla": 'pedidoscli/accion1',
                                    "grupo": 'pedidoscli',
                                    "descripcion": 'Puede ejecutar la accion 1 de pedidos de cliente. Puede ejecutar la accion 1.'
                                },
                                {
                                    "idregla": 'pedidoscli/accion2',
                                    "grupo": 'pedidoscli',
                                    "descripcion": 'Puede ejecutar la accion 2 de pedidos de cliente'
                                }
                            ]
                        );
                    }
                    function getCrudRules(grupo, descripcion)
                    {
                        var crudRules = [
                            {
                                'id': 'get',
                                'desc': 'Puede recibir'
                            },
                            {
                                'id': 'post',
                                'desc': 'Puede crear'
                            },
                            {
                                'id': 'patch',
                                'desc': 'Puede modificar'
                            },
                            {
                                'id': 'delete',
                                'desc': 'Puede eliminar'
                            }
                        ]
                        var rules = [{
                            'idregla': grupo,
                            'grupo': grupo,
                            'descripcion': descripcion
                        }]
                        for (var i = 0; i < crudRules.length; i++) {
                            rules.push({
                                'idregla': grupo + '/' + crudRules[i].id,
                                'grupo': grupo,
                                'descripcion': crudRules[i].desc + ' ' + descripcion.toLowerCase()
                            })
                        }
                        return rules;
                    }
                    """

        cadena_result = qs2py(codigo)
        ok_result = """def getRules(self):
    return qsa.Array().concat(
        self.getCrudRules("pedidoscli", "Pedidos de cliente"),
        qsa.Array(
            [
                qsa.AttributeDict(
                    {
                        "idregla": ("pedidoscli/accion1"),
                        "grupo": ("pedidoscli"),
                        "descripcion": (
                            "Puede ejecutar la accion 1 de pedidos de cliente. Puede ejecutar la accion 1."
                        ),
                    }
                ),
                qsa.AttributeDict(
                    {
                        "idregla": ("pedidoscli/accion2"),
                        "grupo": ("pedidoscli"),
                        "descripcion": ("Puede ejecutar la accion 2 de pedidos de cliente"),
                    }
                ),
            ]
        ),
    )


def getCrudRules(self, grupo, descripcion):
    crudRules: Any = qsa.Array(
        [
            qsa.AttributeDict(
                {
                    "id": ("get"),
                    "desc": ("Puede recibir"),
                }
            ),
            qsa.AttributeDict(
                {
                    "id": ("post"),
                    "desc": ("Puede crear"),
                }
            ),
            qsa.AttributeDict(
                {
                    "id": ("patch"),
                    "desc": ("Puede modificar"),
                }
            ),
            qsa.AttributeDict(
                {
                    "id": ("delete"),
                    "desc": ("Puede eliminar"),
                }
            ),
        ]
    )
    rules: Any = qsa.Array(
        [
            qsa.AttributeDict(
                {
                    "idregla": (grupo),
                    "grupo": (grupo),
                    "descripcion": (descripcion),
                }
            )
        ]
    )
    i: Any = 0
    while_pass = True
    while i < qsa.length(crudRules):
        if not while_pass:
            i += 1
            while_pass = True
            continue
        while_pass = False
        rules.append(
            qsa.AttributeDict(
                {
                    "idregla": (qsa.ustr(grupo, "/", crudRules[i].id)),
                    "grupo": (grupo),
                    "descripcion": (qsa.ustr(crudRules[i].desc, " ", descripcion.lower())),
                }
            )
        )
        i += 1
        while_pass = True
        try:
            i < qsa.length(crudRules)
        except Exception:
            break

    return rules
"""
        self.assertEqual(cadena_result, ok_result)

    def test_execMainScript(self) -> None:
        """Test aqApp.execMainScript."""

        from pineboolib.qsa import qsa

        fun_ = qsa.from_project("sys").iface.execMainScript
        self.assertTrue(fun_ is not None)

    def test_session(self) -> None:
        """Test session conversion."""

        qsa = 'if (sys.interactiveGUI() == "Django") { \nvar sessi = new session();}'
        result = 'if qsa.sys.interactiveGUI() == "Django":\n    sessi: Any = qsa.session()\n'
        self.assertEqual(qs2py(qsa), result)

        qsa2 = 'if (sys.interactiveGUI() == "Django") { \nvar session = new session();}'
        result2 = 'if qsa.sys.interactiveGUI() == "Django":\n    session: Any = qsa.session()\n'
        self.assertEqual(qs2py(qsa2), result2)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Qmainwindow module."""

# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets, QtCore  # type: ignore[import]
from typing import Optional, cast


class QMainWindow(QtWidgets.QMainWindow):
    """QMainWindow class."""

    def child(self, child_name: str, obj: QtCore.QObject) -> Optional[QtWidgets.QWidget]:
        """Return a child especified by name."""

        return cast(QtWidgets.QWidget, self.findChild(QtWidgets.QWidget, child_name))



# -*- coding: utf-8 -*-
"""FLAction Module."""

from pineboolib.application.metadata import pnaction


class FLAction(pnaction.PNAction):
    """
    FLAction Class.

    This class contains information on actions to open forms.

    It is used to automatically link forms with your script,
    interface and source table.

    @author InfoSiAL S.L.
    """



"""
Module for Date type.
"""
from typing import Union, Optional, Any
from PyQt6 import QtCore  # type: ignore[import]
from pineboolib.application.utils.date_conversion import date_dma_to_amd
import datetime


class Date(object):
    """
    Case que gestiona un objeto tipo Date.
    """

    date_: "QtCore.QDate"
    time_: "QtCore.QTime"

    def __init__(
        self, *args: Union["Date", "QtCore.QDate", str, "QtCore.QTime", int, "datetime.date"]
    ) -> None:
        """Create new Date object."""
        super(Date, self).__init__()
        if not args:
            self.date_ = QtCore.QDate.currentDate()
            self.time_ = QtCore.QTime.currentTime()
        elif len(args) <= 2:
            date_ = args[0]
            format_ = args[1] if len(args) == 2 else "yyyy-MM-dd"
            if not isinstance(format_, str):
                raise ValueError("format must be string")
            self.time_ = QtCore.QTime(0, 0)

            if isinstance(date_, str):
                if len(date_) == 10:
                    tmp = date_.split("-")
                    if len(tmp[2]) == 4:
                        date_amd = date_dma_to_amd(date_)
                        if date_amd is None:
                            raise ValueError("Date %s is invalid" % date_)
                        date_ = date_amd

                    self.date_ = QtCore.QDate.fromString(date_, format_)
                else:
                    self.date_ = QtCore.QDate.fromString(date_[0:10], format_)
                    self.time_ = QtCore.QTime.fromString(date_[11:], "hh:mm:ss")

            elif isinstance(date_, Date):
                self.date_ = date_.date_
                self.time_ = date_.time_

            elif isinstance(date_, QtCore.QDate):
                self.date_ = date_
            elif isinstance(date_, (float, int)):
                date_time = QtCore.QDateTime()
                date_time.setMSecsSinceEpoch(int(date_))
                self.date_ = date_time.date()
                self.time_ = date_time.time()
            elif isinstance(date_, (datetime.date)):
                date_ = date_.strftime("%Y-%m-%d")
                self.date_ = QtCore.QDate.fromString(date_, format_)
            else:
                raise ValueError("Unexpected type %s" % type(date_))

            if not self.time_:
                self.time_ = QtCore.QTime(0, 0)
        else:
            year, month, day = args[0], args[1], args[2]
            if not isinstance(year, int) or not isinstance(month, int) or not isinstance(day, int):
                raise ValueError("Expected year, month, day as integers")
            self.date_ = QtCore.QDate(year, month, day)
            self.time_ = QtCore.QTime(0, 0)

    def toString(self, pattern: Optional[str] = None) -> str:
        """
        Return string with date & time data.

        @return cadena de texto con los datos de fecha y hora
        """
        if not pattern:
            pattern = "yyyy-MM-ddT%s" % self.time_.toString("hh:mm:ss")

        return self.date_.toString(pattern)

    def getTime(self) -> int:
        """Get integer representing date & time."""

        return int(self.date_.toString("yyyyMMdd%s" % self.time_.toString("hhmmss")))

    def getYear(self) -> int:
        """
        Return year from date.

        @return año
        """
        return self.date_.year()

    def setYear(self, year: Union[str, int]) -> "Date":
        """
        Set year into current date.

        @param yyyy. Año a setear
        """

        self.date_ = QtCore.QDate.fromString(self.date_.toString("%s-MM-dd" % year), "yyyy-MM-dd")

        return self

    def getMonth(self) -> int:
        """
        Get month as a number from current date.

        @return mes
        """
        return self.date_.month()

    def setMonth(self, month: Union[str, int]) -> "Date":
        """
        Set month into current date.

        @param mm. Mes a setear
        """
        month = str(month)
        if len(month) < 2:
            month = "0%s" % month

        self.date_ = QtCore.QDate.fromString(
            self.date_.toString("yyyy-%s-dd" % month), "yyyy-MM-dd"
        )

        return self

    def getDay(self) -> int:
        """
        Get day from current date.

        @return día
        """
        return self.date_.day()

    def setDay(self, day: Union[str, int]) -> "Date":
        """
        Set given day.

        @param dd. Dia a setear
        """
        day = str(day)

        if len(day) < 2:
            day = "0%s" % day

        self.date_ = QtCore.QDate.fromString(self.date_.toString("yyyy-MM-%s" % day), "yyyy-MM-dd")

        return self

    def getHours(self) -> int:
        """
        Get Hour from Date.

        @return horas
        """
        return self.time_.hour()

    def getMinutes(self) -> int:
        """
        Get Minutes from Date.

        @return minutos
        """
        return self.time_.minute()

    def getSeconds(self) -> int:
        """
        Get Seconds from Date.

        @return segundos
        """
        return self.time_.second()

    def getMilliseconds(self) -> int:
        """
        Get Milliseconds from Date.

        @return milisegundos
        """
        return self.time_.msec()

    getDate = getDay
    # setDate = setDay

    def setDate(self, date: Any) -> "Date":
        """
        Set Date from any format.

        @param date. Fecha a setear
        """
        year_ = self.date_.toString("yyyy")
        month_ = self.date_.toString("MM")
        day_ = str(date)
        if len(day_) == 1:
            day_ = "0" + day_

        str_ = "%s-%s-%s" % (year_, month_, day_)
        self.date_ = QtCore.QDate.fromString(str_, "yyyy-MM-dd")

        return self

    def addDays(self, days: int) -> "Date":
        """
        Return result of adding a particular amount of days to current date.

        @param d. Dias a sumar (o restar) a la fecha dada
        @return nueva fecha calculada
        """
        return Date(self.date_.addDays(days).toString("yyyy-MM-dd"))

    def addMonths(self, months: int) -> "Date":
        """
        Return result of adding given number of months to this date.

        @param m. Meses a sumar (o restar) a la fecha dada
        @return nueva fecha calculada
        """
        return Date(self.date_.addMonths(months).toString("yyyy-MM-dd"))

    def addYears(self, years: int) -> "Date":
        """
        Return result of adding given number of years to this date.

        @param y. Años a sumar (o restar) a la fecha dada
        @return nueva fecha calculada
        """
        return Date(self.date_.addYears(years).toString("yyyy-MM-dd"))

    @classmethod
    def parse(cls, value: str) -> float:
        """Parse a ISO string into a date."""
        # return Date(value, "yyyy-MM-dd")
        if "T" not in value:
            value = "%sT00:00:00" % value
        return datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S").timestamp() * 1000

    def __str__(self) -> str:
        """Support for str()."""
        return self.toString()

    def __repr__(self) -> str:
        """Support for str()."""
        return self.toString()

    def __lt__(self, other: Union[str, "Date"]) -> bool:
        """Support for comparisons."""
        return str(self) < str(other)

    def __le__(self, other: Union[str, "Date"]) -> bool:
        """Support for comparisons."""
        return str(self) <= str(other)

    def __ge__(self, other: Union[str, "Date"]) -> bool:
        """Support for comparisons."""
        return str(self) >= str(other)

    def __gt__(self, other: Union[str, "Date"]) -> bool:
        """Support for comparisons."""
        return str(self) > str(other)

    def __eq__(self, other: Any) -> bool:
        """Support for comparisons."""
        return str(other) == str(self)

    def __ne__(self, other: Any) -> bool:
        """Support for comparisons."""
        return not self.__eq__(other)



"""
Convert a date to different formats.
"""


from PyQt6 import QtCore  # type: ignore

from pineboolib.application.qsatypes import date
import datetime
import re
from typing import Optional, List, Union


def date_dma_to_amd(date_str: str) -> Optional[str]:
    """
    Convert day, month, year to year, month day.

    @param date_str: date string.
    @return date string formated.
    """

    if not date_str:
        return None

    date_str = str(date_str)
    if date_str.find("T") > -1:
        date_str = date_str[: date_str.find("T")]

    dia_ = None
    mes_ = None
    ano_ = None

    array_: List[str] = re.split("-|/", date_str)

    if len(array_) > 1:
        if len(array_) == 3:
            dia_, mes_, ano_ = array_ if len(array_[0]) == 2 else reversed(array_)
    elif len(date_str) > 7:
        dia_ = date_str[0:2]
        mes_ = date_str[2:2]
        ano_ = date_str[4:4]

    return "%s-%s-%s" % (ano_, mes_, dia_) if ano_ and dia_ and mes_ else ""


def date_amd_to_dma(date_str: str) -> str:
    """
    Convert year, month day to day, month, year.

    @param date_str: date string.
    @return date string formated.
    """

    if not date_str:
        return ""

    date_str = str(date_str)
    if date_str.find("T") > -1:
        date_str = date_str[: date_str.find("T")]

    dia_ = None
    mes_ = None
    ano_ = None

    array_: List[str] = re.split("-|/", date_str)

    if len(array_) > 1:
        if len(array_) == 3:
            dia_, mes_, ano_ = reversed(array_) if len(array_[0]) == 4 else array_
    elif len(date_str) > 7:
        ano_ = date_str[0:4]
        mes_ = date_str[4:2]
        dia_ = date_str[6:2]

    return "%s-%s-%s" % (dia_, mes_, ano_) if ano_ and dia_ and mes_ else ""


def convert_to_qdate(
    value: Union["datetime.date", "date.Date", str, "QtCore.QDate"]
) -> "QtCore.QDate":
    """
    Convert different date formats to QDate.

    @param date: Date to convert.
    @return QDate with the value of the given date.
    """

    if isinstance(value, date.Date):
        value = value.toString()  # QDate -> str
    elif isinstance(value, datetime.date):
        value = str(value)

    if isinstance(value, str):
        if "T" in value:
            value = value[: value.find("T")]

        new_value = date_amd_to_dma(value) if len(value.split("-")[0]) == 4 else value
        value = QtCore.QDate.fromString(new_value, "dd-MM-yyyy")

    return value



"""Spinbox module."""

from PyQt6 import QtWidgets  # type: ignore[import]


class SpinBox(QtWidgets.QWidget):
    """SpinBox class."""

    _label: "QtWidgets.QLabel"
    _spin_box: "QtWidgets.QSpinBox"
    _lay: "QtWidgets.QHBoxLayout"

    def __init__(self, parent=None):
        """Initialize."""

        super().__init__(parent)

        self._lay = QtWidgets.QHBoxLayout(self)
        self._label = QtWidgets.QLabel()
        self._spin_box = QtWidgets.QSpinBox()

        self._lay.addWidget(self._label)
        self._lay.addWidget(self._spin_box)

    def getMax(self) -> int:
        """Return Maximum."""
        return self._spin_box.maximum()

    def setMax(self, max: int) -> None:
        """Set Maximum."""
        self._spin_box.setMaximum(max)

    def getMin(self) -> int:
        """Return Minimum."""
        return self._spin_box.minimum()

    def setMin(self, min: int) -> None:
        """Set Minimum."""
        self._spin_box.setMinimum(min)

    def getValue(self) -> int:
        """Return value."""
        return self._spin_box.value()

    def setValue(self, value: int) -> None:
        """Set Minimum."""
        self._spin_box.setValue(value)

    def getLabel(self) -> str:
        """Return label."""

        return self._label.text()

    def setLabel(self, label: str) -> None:
        """Set label."""

        self._label.setText(label)

    maximum: int = property(getMax, setMax)  # type: ignore [assignment] # noqa: F821
    minimum: int = property(getMin, setMin)  # type: ignore [assignment] # noqa: F821
    value: int = property(getValue, setValue)  # type: ignore [assignment] # noqa: F821
    label: str = property(getLabel, setLabel)  # type: ignore [assignment] # noqa: F821



# -*- coding: utf-8 -*-
"""
KUT2FPDF module.

Creates a pdf representation of a KUT
"""
import datetime
import re
import os
from xml.etree.ElementTree import Element
from pineboolib import application
from pineboolib import logging
from pineboolib.core.utils.utils_base import load2xml
from pineboolib.application.utils.check_dependencies import check_dependencies
from pineboolib.application.parsers.parser_kut import kparsertools
from pineboolib.core import settings
from pineboolib.application import qsadictmodules, connections


from typing import Any, Optional, Union, List, Dict, TYPE_CHECKING

if TYPE_CHECKING:
    from fpdf import FPDF  # type: ignore # pragma: no cover

LOGGER = logging.get_logger(__name__)


class Kut2FPDF(object):
    """
    Convert kuts to pyFPDF.
    """

    _document: "FPDF"

    _xml: Element
    _xml_data: Element
    _page_orientation: str
    _page_size: List[int]
    _bottom_margin: int
    _left_margin: int
    _right_margin: int
    _top_margin: int
    _page_top: Dict[int, int]
    _data_row: Element
    _parser_tools: "kparsertools.KParserTools"
    _avalible_fonts: List[str]
    _unavalible_fonts: List[str]
    design_mode: bool
    _actual_data_line: Optional["Element"]
    _no_print_footer: bool
    _actual_section_size: int
    increase_section_size: int
    last_detail: bool
    actual_data_level: int
    last_data_processed: "Element"
    prev_level: int
    draws_at_header: Dict[str, str]
    detailn: Dict[str, int]
    name_: str
    _actual_append_page_no: int
    reset_page_count: bool
    last_register: bool

    def __init__(self) -> None:
        """Initialize."""

        check_dependencies({"fpdf": "fpdf2"})

        self._parser_tools = kparsertools.KParserTools()
        self._avalible_fonts = []
        self._page_top: Dict[int, int] = {}
        self._unavalible_fonts = []
        self.design_mode = settings.CONFIG.value("ebcomportamiento/kugar_debug_mode", False)
        self._actual_data_line = None
        self._no_print_footer = False
        self.increase_section_size = 0
        self.actual_data_level = 0
        self.prev_level = -1
        self.draws_at_header = {}
        self.detailn = {}
        self.name_ = ""
        self._actual_append_page_no = -1
        self.reset_page_count = False
        self.new_page = False
        self._document = None
        self.last_register = False

    def parse(
        self, name: str, kut: str, data: str, report: "FPDF" = None, flags: List[int] = []
    ) -> Optional[str]:
        """
        Parse string containing ".kut" file into a pdf and return its file path.

        @param name. Filename path for ".kut".
        @param kut. String with ".kut" file contents.
        @param data. String with data to be used in the report.
        @return Path to PDF file.
        """
        try:
            self._xml = self._parser_tools.loadKut(kut).getroot()
        except Exception:
            LOGGER.exception("KUT2FPDF: Problema al procesar %s.kut", name)
            return None
        try:
            self._xml_data = load2xml(data).getroot()
        except Exception:
            LOGGER.exception("KUT2FPDF: Problema al procesar xml_data")
            return None

        application.PROJECT.message_manager().send(
            "progress_dialog_manager", "create", ["Pineboo", len(self._xml_data), "kugar"]
        )
        application.PROJECT.message_manager().send(
            "progress_dialog_manager", "setLabelText", ["Creando informe ...", "kugar"]
        )

        self.name_ = name
        self.setPageFormat(self._xml)
        # self._page_orientation =
        # self._page_size =
        if report is None:
            import fpdf

            self._actual_append_page_no = 0
            self._document = fpdf.FPDF(self._page_orientation, "pt", self._page_size)
            for font in self._document.core_fonts:
                LOGGER.debug("KUT2FPDF :: Adding font %s", font)
                self._avalible_fonts.append(font)
        else:
            self._document = report
        # Seteamos rutas a carpetas con tipos de letra ...

        # Cargamos las fuentes disponibles
        next_page_break = (flags[2] == 1) if len(flags) == 3 else True
        page_append = (flags[1] == 1) if len(flags) > 1 else False
        page_display = (flags[0] == 1) if len(flags) > 0 else False

        if page_append:
            self.prev_level = -1
            self.last_detail = False

        page_break = False
        if self.new_page:
            page_break = True
            self.new_page = False

        if self.reset_page_count:
            self.reset_page_no()
            self.reset_page_count = False

        if self.design_mode:
            LOGGER.warning("Append %s", page_append)
            LOGGER.warning("Display %s", page_display)
            LOGGER.warning("Page break %s", next_page_break)

        if next_page_break:
            self.reset_page_count = True

        if page_display:
            self.new_page = True

        self.processDetails(not page_break)

        # FIXME:Alguno valores no se encuentran
        for pages in self._document.pages.keys():
            page_content = self._document.pages[pages]["content"]
            for header in self.draws_at_header.keys():
                page_content = page_content.replace(header, str(self.draws_at_header[header]))

            self._document.pages[pages]["content"] = page_content

        self._document.set_title(self.name_)
        self._document.set_author("Pineboo - kut2fpdf plugin")

        return self._document

    def get_file_name(self) -> Optional[str]:
        """Retrieve file name where PDF should be saved."""
        import os

        pdf_name = application.PROJECT.tmpdir
        pdf_name += "/%s_%s.pdf" % (self.name_, datetime.datetime.now().strftime("%Y%m%d%H%M%S"))
        if os.path.exists(pdf_name):
            os.remove(pdf_name)
        if self._document is not None:
            self._document.output(pdf_name)
            return pdf_name
        else:
            return None

    def topSection(self) -> int:
        """
        Retrieve top section margin for current page to calculate object positions.

        @return Number with ceiling for current page.
        """
        return self._page_top[int(self._document.page_no())]

    def setTopSection(self, value: int) -> None:
        """
        Update top section for current page.

        Usually updated when processing a section.
        @param value. Number specifying new ceiling.
        """
        self._actual_section_size = value - self.topSection()
        self._page_top[int(self._document.page_no())] = value

    def newPage(self, data_level: int, add_on_header: bool = True) -> None:
        """
        Add a new page to the document.
        """
        self._document.add_page(self._page_orientation)
        self._page_top[int(self._document.page_no())] = self._top_margin
        self._document.set_margins(
            self._left_margin, self._top_margin, self._right_margin
        )  # Lo dejo pero no se nota nada
        self._no_print_footer = False
        if self.design_mode:
            self.draw_margins()

        self._actual_section_size = 0
        self._actual_append_page_no += 1
        if self.design_mode:
            LOGGER.warning("Nueva página %s", self.number_pages())

        pg_headers = self._xml.findall("PageHeader")

        for page_header in pg_headers:
            if self.number_pages() == 1 or page_header.get("PrintFrequency") == "1":
                ph_level = (
                    page_header.get("Level") if page_header.get("Level") is not None else None
                )
                self.processSection("PageHeader", int(ph_level or "0"))
                break

        if add_on_header and not self.number_pages() == 1:
            for level in range(data_level + 1):
                self.processSection("AddOnHeader", int(level))

        # Por ahora se omite detail header

    def processDetails(self, keep_page: Optional[bool] = None) -> None:
        """
        Process detail secions with their matching detailHeader and detailFooter.
        """
        # Procesamos la cabecera si procede ..

        top_level = 0
        level = 0
        first_page_created = (
            keep_page if keep_page is not None and self._document.page_no() > 0 else False
        )

        rows_array = self._xml_data.findall("Row")
        i = 0

        for data in rows_array:
            self.last_detail = False
            self._actual_data_line = data
            level_str: str = data.get("level") or "0"

            level = int(level_str)
            if level > top_level:
                top_level = level

            if data is not rows_array[len(rows_array) - 1]:
                next_data = rows_array[rows_array.index(data) + 1]
                next_level = int(next_data.get("level") or 0)
                if int(top_level) > next_level:
                    self.last_detail = True

            if not first_page_created:
                self.newPage(level)

                first_page_created = True

            if rows_array[len(rows_array) - 1] is data:
                self.last_register = True
                self.last_detail = True

            if level < self.prev_level:
                for lev in range(level + 1, self.prev_level + 1):
                    self.processData("DetailFooter", self.last_data_processed, lev)

            if not str(level) in self.detailn.keys():
                self.detailn[str(level)] = 0
            else:
                self.detailn[str(level)] += 1

            if level > self.prev_level:
                self.processData("DetailHeader", data, level)

            self.processData("Detail", data, level)

            self.last_data_processed = data

            self.prev_level = level

            application.PROJECT.message_manager().send(
                "progress_dialog_manager", "setProgress", [i, "kugar"]
            )
            i += 1

        if not self._no_print_footer and hasattr(self, "last_data_processed"):
            for lev in reversed(range(top_level + 1)):
                self.processData("DetailFooter", self.last_data_processed, lev)

        application.PROJECT.message_manager().send("progress_dialog_manager", "destroy", ["kugar"])

    def processData(self, section_name: str, data: "Element", data_level: int) -> None:
        """
        Check if detailHeader + detail + detailFooter do fit in the remaining page and create a new page if not.

        @param section_name. Section name to check
        @param data. Data to check
        @param data_level. Section level
        """
        self.actual_data_level = data_level
        list_sections = self._xml.findall(section_name)
        # data_size = len(listDF)

        for section in list_sections:
            draw_if = section.get("DrawIf")
            show = True
            if draw_if:
                show = data.get(draw_if) not in ("", "False", "None", False)
            if section.get("Level") == str(data_level) and show:
                current_size = self._parser_tools.getHeight(section)
                page_footer_size = 0
                detail_footer_size = 0
                add_on_footer_size = 0
                if section_name in ("DetailHeader", "Detail"):
                    if section_name == "DetailHeader":
                        for detail in self._xml.findall("Detail"):
                            if detail.get("Level") == str(data_level):
                                current_size += self._parser_tools.getHeight(detail)
                                break

                    page_footer: Any = self._xml.get("PageFooter")
                    if isinstance(page_footer, Element):
                        if (
                            self._document.page_no() == 1
                            or page_footer.get("PrintFrecuency") == "1"
                        ):
                            page_footer_size += self._parser_tools.getHeight(page_footer)

                    for detail_footer in self._xml.findall("DetailFooter"):
                        if detail_footer.get("Level") or "0" <= str(data_level):
                            detail_footer_size = self._parser_tools.getHeight(detail_footer)
                            break

                for add_footer in self._xml.findall("AddOnFooter"):
                    if add_footer.get("Level") == str(data_level):
                        add_on_footer_size += self._parser_tools.getHeight(add_footer)
                        break

                limit_one = self._document.h + 5  # falta altura siguente suma_importes

                if section_name == "Detail":
                    limit_one -= detail_footer_size
                else:
                    limit_one -= self._bottom_margin

                limit_two = (
                    self._document.h - self._bottom_margin - page_footer_size - add_on_footer_size
                )

                virtual_size = self.topSection() + current_size  # top_size + current_size

                if self.design_mode:
                    LOGGER.warning(
                        "virtual name: %s, size: %s (%s + %s), l1: %s, l2: %s",
                        section_name,
                        virtual_size,
                        self.topSection(),
                        current_size,
                        limit_one,
                        limit_two,
                    )

                if virtual_size > limit_one and section_name != "DetailFooter":  # Si nos pasamos
                    if self.design_mode:
                        LOGGER.warning("ME PASO con %s!", section_name)
                    self._no_print_footer = True

                    if virtual_size >= limit_two or self.last_detail:
                        self.last_detail = False
                        self.processSection("AddOnFooter", int(data_level))

                        self.newPage(data_level)

                self.processXML(section, data)

                if (
                    section.get("NewPage") == "true"
                    and not self.last_register
                    and self.topSection() > self._top_margin
                ):
                    self.newPage(data_level, False)

                break  # Se ejecuta una sola instancia

    def processSection(self, name: str, level: int = 0) -> None:
        """
        Process non-detail sections.

        @param name. Section name to process
        """
        sec_list = self._xml.findall(name)
        sec_ = None
        for section in sec_list:
            if section.get("Level") == str(level) or section.get("Level") is None:
                sec_ = section
                break

        if sec_ is not None:
            if (
                sec_.get("PrintFrequency") == "1"
                or self._document.page_no() == 1
                or name in ("AddOnHeader", "AddOnFooter")
            ):
                self.processXML(sec_)

    def processXML(self, xml: "Element", data: Optional["Element"] = None) -> None:
        """
        Process single XML element.

        @param xml: Element to process
        @param. data: Line affected
        """

        fix_height = True
        if data is None:
            data = self._actual_data_line

        if self.design_mode and data is not None:
            LOGGER.warning(
                "Procesando: %s level: %s size: %s", xml.tag, data.get("level"), xml.get("Height")
            )

        size_updated = False
        if xml.tag == "DetailFooter":
            if xml.get("PlaceAtBottom") == "true":
                height = self._parser_tools.getHeight(xml)
                self.setTopSection(self._document.h - height - self.increase_section_size)
                size_updated = True

        if xml.tag == "PageFooter":
            fix_height = False

        if not size_updated:
            self.fix_extra_size()  # Sirve para actualizar la altura con lineas que se han partido porque son muy largas

        for label in xml.iter("Label"):
            self.processText(label, data, fix_height)

        for field in xml.iter("Field"):
            self.processText(field, data, fix_height)

        for special in xml.iter("Special"):
            self.processText(special, data, fix_height, xml.tag)

        for calculated in xml.iter("CalculatedField"):
            self.processText(calculated, data, fix_height, xml.tag)

        # Busco draw_at_header en DetailFooter y los meto también
        if xml.tag == "DetailHeader":
            detail_level = xml.get("Level")
            if detail_level is None:
                raise Exception("Level tag not found")
            for detail_footer in self._xml.iter("DetailFooter"):
                if detail_footer.get("Level") == detail_level:
                    for calculated_filed in detail_footer.iter("CalculatedField"):
                        if calculated_filed.get("DrawAtHeader") == "true":
                            header_name = "%s_header_%s_%s" % (
                                self.detailn[detail_level],
                                detail_level,
                                calculated_filed.get("Field"),
                            )
                            self.draws_at_header[header_name] = ""
                            self.processText(calculated_filed, data, fix_height, xml.tag)

        for line in xml.iter("Line"):
            self.processLine(line, fix_height)

        if not size_updated:
            self.setTopSection(self.topSection() + self._parser_tools.getHeight(xml))

        self.increase_section_size = 0

    def fix_extra_size(self) -> None:
        """Increase size of the section if needed."""
        if self.increase_section_size > 0:
            if self.design_mode:
                LOGGER.warning("Increasing extra_size %s", self.increase_section_size)
            self.setTopSection(self.topSection() + self.increase_section_size)
            self.increase_section_size = 0

    def processLine(self, xml: "Element", fix_height: bool = True) -> None:
        """
        Process single line.

        @param xml. Sección de xml a procesar.
        @param fix_height. Ajusta la altura a los .kut originales, excepto el pageFooter.
        """

        color = xml.get("Color")
        red = 0 if not color else int(color.split(",")[0])
        green = 0 if not color else int(color.split(",")[1])
        blue = 0 if not color else int(color.split(",")[2])

        style = int(xml.get("Style") or "0")
        width = int(xml.get("Width") or "0")
        pos_x1 = self.calculateLeftStart(xml.get("X1") or "0")
        pos_x1 = self.calculateWidth(pos_x1, 0, False)
        pos_x2 = self.calculateLeftStart(xml.get("X2") or "0")
        pos_x2 = self.calculateWidth(pos_x2, 0, False)

        # Ajustar altura a secciones ya creadas
        pos_y1 = int(xml.get("Y1") or "0") + self.topSection()
        pos_y2 = int(xml.get("Y2") or "0") + self.topSection()
        if fix_height:
            pos_y1 = self._parser_tools.ratio_correction_h(pos_y1)
            pos_y2 = self._parser_tools.ratio_correction_h(pos_y2)

        self._document.set_line_width(self._parser_tools.ratio_correction_h(width))
        self._document.set_draw_color(red, green, blue)
        dash_length = 1
        space_length = 1
        if style == 2:
            dash_length = 20
            space_length = 20
        elif style == 3:
            dash_length = 10
            space_length = 10

        self._document.dashed_line(pos_x1, pos_y1, pos_x2, pos_y2, dash_length, space_length)
        # else:
        #    self._document.line(X1, Y1, X2, Y2)

    def calculateLeftStart(self, position: Union[str, int, float]) -> int:
        """
        Check if left margin is exceeded for current page.

        @param position. Position to check.
        @return Revised position.
        """
        return self._parser_tools.ratio_correction_w(int(position)) + self._left_margin

    def calculateWidth(self, width: int, pos_x: int, fix_ratio: bool = True) -> int:
        """
        Check if right margin is exceeded for current page.

        @param x. Position to check.
        @return Revised position.
        """
        limit = self._document.w - self._right_margin
        ret_: int

        if fix_ratio:
            width = self._parser_tools.ratio_correction_w(width)
            pos_x = self._parser_tools.ratio_correction_w(pos_x)
            ret_ = width
            if pos_x + width > limit:
                ret_ = limit - pos_x
        else:
            ret_ = width

        return ret_

    def processText(
        self,
        xml: Element,
        data_row: Optional[Element] = None,
        fix_height: bool = True,
        section_name: Optional[str] = None,
    ) -> None:
        """
        Check tag (calculated, label, special or image).

        @param xml. XML section to process.
        @param fix_height. Revise height from original .kut file except pageFooter.
        """
        is_image = False
        is_barcode = False
        text: str = xml.get("Text") or ""
        # borderColor = xml.get("BorderColor")
        field_name = xml.get("Field") or ""
        # x,y,W,H se calcula y corrigen aquí para luego estar correctos en los diferentes destinos posibles
        width = int(xml.get("Width") or "0")

        height = self._parser_tools.getHeight(xml)

        pos_x = int(xml.get("X") or "0")

        pos_y = (
            int(xml.get("Y") or "0") + self.topSection()
        )  # Añade la altura que hay ocupada por otras secciones
        if fix_height:
            pos_y = self._parser_tools.ratio_correction_h(
                pos_y
            )  # Corrige la posición con respecto al kut original

        data_type = xml.get("DataType")

        if xml.tag == "Field" and data_row is not None:
            text = data_row.get(field_name)  # type: ignore [assignment]

        elif xml.tag == "Special":
            alternative_text = text
            if xml.get("Type") == "0":
                text = "Date"
            if xml.get("Type") == "1":
                text = "PageNo"

            text = self._parser_tools.getSpecial(text, self._actual_append_page_no)
            if text == "None":
                text = alternative_text

        calculation_type = xml.get("CalculationType")

        if calculation_type is not None and xml.tag not in ["Field", "Special"]:
            if calculation_type == "6":
                function_name = xml.get("FunctionName")
                if function_name and data_row is not None:
                    try:
                        nodo = self._parser_tools.convertToNode(data_row)
                        function_list = function_name.split(".")
                        func_ = None

                        if len(function_list) == 2:
                            func_ = getattr(
                                qsadictmodules.QSADictModules.from_project(function_list[0]).iface,
                                function_list[1],
                                None,
                            )
                        elif len(function_list) == 3:
                            module_ = getattr(
                                qsadictmodules.QSADictModules.from_project(function_list[0]),
                                function_list[1],
                            )
                            func_ = getattr(module_, function_list[2], None)
                        else:
                            LOGGER.warning("PLEASE FIXME: diferent of 2 (%s)", function_list)
                        if func_ is not None:
                            arguments = [nodo, field_name]
                            expected_arguments = connections.get_expected_args_num(func_)

                            ret_ = func_(*arguments[:expected_arguments])

                            if ret_ is False:
                                return
                            else:
                                text = str(ret_)

                    except Exception:
                        LOGGER.exception("KUT2FPDF:: Error llamando a function %s", function_name)
                        return
                else:
                    return
            elif calculation_type == "1":
                text = self._parser_tools.calculate_sum(
                    field_name, self.last_data_processed, self._xml_data, self.actual_data_level
                )

            elif calculation_type in ("5"):
                if data_row is None:
                    data_row = self._xml_data[0]

                text = data_row.get(field_name) or ""

        if data_type not in [None, ""]:
            text = self._parser_tools.calculated(text, int(data_type), xml, data_row)  # type: ignore [arg-type]

            if data_type == "5":
                is_image = True

            elif data_type == "6":
                is_barcode = True

        if xml.get("BlankZero") == "1" and text is not None:
            res_ = re.findall(r"\d+", text)
            if res_ == ["0"]:
                return

        if text is not None:
            temporal = settings.CONFIG.value("ebcomportamiento/temp_dir")
            if text.startswith(temporal):
                is_image = True

        # negValueColor = xml.get("NegValueColor")
        # Currency = xml.get("Currency")
        #
        # commaSeparator = xml.get("CommaSeparator")
        # dateFormat = xml.get("DateFormat")

        if is_image:
            self.draw_image(pos_x, pos_y, width, height, xml, text)
        elif is_barcode:
            self.draw_barcode(pos_x, pos_y, width, height, xml, text)
        elif data_row is not None:
            level = data_row.get("level") or "0"
            if level and str(level) in self.detailn.keys():
                val = "%s_header_%s_%s" % (self.detailn[str(level)], level, field_name)

            if xml.get("DrawAtHeader") == "true" and level:
                if section_name == "DetailHeader":
                    val = ""
                    self.drawText(pos_x, pos_y, width, height, xml, val)

            if section_name == "DetailFooter" and xml.get("DrawAtHeader") == "true":
                self.draws_at_header[val] = text

            else:
                self.drawText(pos_x, pos_y, width, height, xml, text)

    def drawText(
        self, pos_x: int, pos_y: int, width: int, height: int, xml: "Element", txt: str
    ) -> None:
        """
        Draw a text field onto the page.

        @param pos_x. Label X Pos.
        @param pos_y. Label Y Pos.
        @param width. Label Width.
        @param height. Label Height.
        @param xml. Related XML Section
        @param txt. Computed text of the label to be created.
        """

        if txt in ("None", None):
            # return
            txt = ""

        if width == 0 and height == 0:
            return

        txt = self._parser_tools.restore_text(txt)

        resizeable = False

        if xml.get("ChangeHeight") == "1":
            resizeable = True

        # height_resized = False
        orig_x = pos_x
        orig_y = pos_y
        orig_w = width
        orig_h = height
        # Corregimos margenes:
        pos_x = self.calculateLeftStart(pos_x)
        width = self.calculateWidth(width, pos_x)

        # bg_color = xml.get("BackgroundColor").split(",")
        fg_color = self.get_color(xml.get("ForegroundColor") or "")
        self._document.set_text_color(fg_color[0], fg_color[1], fg_color[2])

        # self._document.set_draw_color(255, 255, 255)

        # if xml.get("BorderStyle") == "1":
        # FIXME: Hay que ajustar los margenes

        # font_name, font_size, font_style
        font_style = ""
        font_size = int(xml.get("FontSize") or "0")
        font_name_orig = (
            (xml.get("FontFamily") or "").lower()
            if xml.get("FontFamily") is not None
            else "helvetica"
        )
        font_name = font_name_orig

        font_w = int(xml.get("FontWeight") or "50")

        if font_w == 50:  # Normal
            font_w = 100
        elif font_w >= 65:
            font_style += "B"
            font_w = 100

        font_italic = xml.get("FontItalic")
        font_underlined = xml.get("FontUnderlined")  # FIXME: hay que ver si es así

        # background_color = self.get_color(xml.get("BackgroundColor"))
        # if background_color != [255,255,255]: #Los textos que llevan fondo no blanco van en negrita
        #    font_style += "B"

        if font_italic == "1":
            font_style += "I"

        if font_underlined == "1":
            font_style += "U"

        font_name = font_name.replace(" narrow", "")

        font_full_name = "%s%s" % (font_name, font_style)

        if font_full_name not in self._avalible_fonts:
            font_found: Optional[str] = None
            if font_full_name not in self._unavalible_fonts:
                font_found = self._parser_tools.find_font(font_full_name, font_style)
            if font_found:
                if self.design_mode:
                    LOGGER.warning(
                        "KUT2FPDF::Añadiendo el tipo de letra %s %s (%s)",
                        font_name,
                        font_style,
                        font_found,
                    )
                self._document.add_font(font_name, font_style, font_found, True)
                self._avalible_fonts.append(font_full_name)

            else:
                if font_full_name not in self._unavalible_fonts:
                    if self.design_mode:
                        LOGGER.warning(
                            "KUT2FPDF:: No se encuentra el tipo de letra %s. Sustituido por helvetica%s."
                            % (font_full_name, font_style)
                        )
                    self._unavalible_fonts.append(font_full_name)
                font_name = "helvetica"

        if font_name is not font_name_orig and font_name_orig.lower().find("narrow") > -1:
            font_w = 85

        self._document.set_font(font_name, font_style, font_size)
        self._document.set_stretching(font_w)
        # Corregir alineación
        vertical_alignment = xml.get("VAlignment")  # 0 izquierda, 1 centrado,2 derecha
        horizontal_alignment = xml.get("HAlignment")

        # layout_direction = xml.get("layoutDirection")

        start_section_size = self._actual_section_size
        result_section_size = 0
        # Miramos si el texto sobrepasa el ancho

        array_text: List[str] = []
        array_n = []
        text_lines = []
        if txt.find("\n") > -1:
            for line_text in txt.split("\n"):
                array_n.append(line_text)
        if array_n:  # Hay saltos de lineas ...
            for line_ in array_n:
                text_lines.append(line_)
        else:  # No hay saltos de lineas
            text_lines.append(txt)

        for text_line in text_lines:
            if len(text_line) > 1:
                if text_line[0] == " " and text_line[1] != " ":
                    text_line = text_line[1:]
            str_width = self._document.get_string_width(text_line)
            if (
                str_width > width - 10 and xml.tag != "Label" and resizeable
            ):  # Una linea es mas larga que el ancho del campo(Le quito 10 al ancho maximo para que corte parecido a kugar original)
                # height_resized = True
                array_text = self.split_text(text_line, width - 10)
            else:
                array_text.append(text_line)

        # calculated_h = orig_h * len(array_text)
        self.drawRect(orig_x, orig_y, orig_w, orig_h, xml)

        processed_lines = 0
        extra_size = 0
        for actual_text in array_text:
            processed_lines += 1

            if processed_lines > 1:
                if self.design_mode:
                    LOGGER.warning("EXTRA SIZE ! %s", array_text)
                extra_size += font_size + 2

            if horizontal_alignment == "1":  # sobre X
                # Centrado
                pos_x = pos_x + (width / 2) - (self._document.get_string_width(actual_text) / 2)
                # x = x + (width / 2) - (str_width if not height_resized else width / 2)
            elif horizontal_alignment == "2":
                # Derecha
                pos_x = (
                    pos_x + width - self._document.get_string_width(actual_text) - 2
                )  # -2 de margen
                # x = x + width - str_width if not height_resized else width
            else:
                # Izquierda
                if processed_lines == 1:
                    pos_x = pos_x + 2

            if vertical_alignment == "1":  # sobre Y
                # Centrado
                # y = (y + ((H / 2) / processed_lines)) + (((self._document.font_size_pt / 2) / 2) * processed_lines)
                pos_y = int((orig_y + (orig_h / 2)) + ((self._document.font_size_pt / 2) / 2))

                if len(array_text) > 1:
                    pos_y = pos_y - (font_size // 2)

            elif vertical_alignment == "2":
                # Abajo
                pos_y = orig_y + orig_h - font_size
            else:
                # Arriba
                pos_y = orig_y + font_size

            pos_y = pos_y + extra_size

            if self.design_mode:
                self.write_debug(
                    self.calculateLeftStart(orig_x),
                    pos_y,
                    "Hal:%s, Val:%s, T:%s st:%s"
                    % (horizontal_alignment, vertical_alignment, txt, font_w),
                    6,
                    "green",
                )
                if xml.tag == "CalculatedField":
                    self.write_debug(
                        self.calculateLeftStart(orig_x),
                        pos_y,
                        "CalculatedField:%s, Field:%s"
                        % (xml.get("FunctionName"), xml.get("Field")),
                        3,
                        "blue",
                    )

            if self.design_mode:
                """LOGGER.warning(
                    "draw ! %s, x: %s, y: %s, ox:%s, oy:%s",
                    actual_text,
                    pos_x,
                    pos_y,
                    orig_x,
                    orig_y,
                )"""
            self._document.text(pos_x, pos_y, actual_text)
            result_section_size += start_section_size

        result_section_size = result_section_size - start_section_size

        if (
            self.increase_section_size < extra_size
        ):  # Si algun incremento extra hay superior se respeta
            self.increase_section_size = extra_size

    def split_text(self, texto: str, limit_w: int) -> List[str]:
        """Split text into lines based on visual width."""
        list_ = []
        linea_: Optional[str] = None

        for parte in texto.split(" "):
            if linea_ is None and parte == "":
                continue

            if linea_ is not None:
                if self._document.get_string_width(linea_ + parte) > limit_w:
                    list_.append(linea_)
                    linea_ = ""
            else:
                linea_ = ""

            linea_ += "%s " % parte
        if linea_ is not None:
            list_.append(linea_)
        return list_

    def get_color(self, value: str) -> List[int]:
        """Convert color text into [r,g,b] array."""
        lvalue = value.split(",")
        red: int
        green: int
        blue: int
        if len(lvalue) == 3:
            red = int(lvalue[0])
            green = int(lvalue[1])
            blue = int(lvalue[2])
        else:
            red = int(value[0:2])
            green = int(value[3:5])
            blue = int(value[6:8])

        return [red, green, blue]

    def drawRect(
        self, pos_x: int, pos_y: int, width: int, height: int, xml: Optional["Element"] = None
    ) -> None:
        """
        Draw a rectangle in current page.

        @param pos_x. left side
        @param pos_y. top side
        @param width. width
        @param height. heigth
        """
        style_ = ""
        border_color = None
        bg_color = None
        line_width = self._document.line_width
        border_width = 0.2
        # Calculamos borde  y restamos del ancho
        orig_x = pos_x
        orig_y = pos_y
        orig_w = width

        pos_x = self.calculateLeftStart(orig_x)
        width = self.calculateWidth(width, pos_x)

        if xml is not None and not self.design_mode:
            if xml.get("BorderStyle") == "1":
                border_color = self.get_color(xml.get("BorderColor") or "")
                self._document.set_draw_color(border_color[0], border_color[1], border_color[2])
                style_ += "D"

            bg_color = self.get_color(xml.get("BackgroundColor") or "")
            self._document.set_fill_color(bg_color[0], bg_color[1], bg_color[2])
            style_ = "F" + style_

            border_width = int(xml.get("BorderWidth") or "0" if xml.get("BorderWidth") else 0.2)
        else:
            self.write_cords_debug(pos_x, pos_y, width, height, orig_x, orig_w)
            style_ = "D"
            self._document.set_draw_color(0, 0, 0)

        if style_ != "":
            self._document.set_line_width(border_width)

            self._document.rect(pos_x, pos_y, width, height, style_)
            self._document.set_line_width(line_width)

            self._document.set_xy(orig_x, orig_y)
        # self._document.set_draw_color(255, 255, 255)
        # self._document.set_fill_color(0, 0, 0)

    def write_cords_debug(
        self,
        pos_x: Union[float, int],
        pos_y: Union[float, int],
        width: Union[float, int],
        height: Union[float, int],
        orig_x: Union[float, int],
        orig_w: Union[float, int],
    ) -> None:
        """Debug for Kut coordinated."""
        self.write_debug(
            int(pos_x),
            int(pos_y),
            "X:%s Y:%s W:%s H:%s orig_x:%s, orig_W:%s"
            % (
                round(pos_x, 2),
                round(pos_y, 2),
                round(width, 2),
                round(height, 2),
                round(orig_x, 2),
                round(orig_w, 2),
            ),
            2,
            "red",
        )

    def write_debug(
        self, pos_x: int, pos_y: int, text: str, height: int, color: Optional[str] = None
    ) -> None:
        """Write debug data into the report."""
        orig_color = self._document.text_color
        red = 0
        green = 0
        blue = 0
        current_font_family = self._document.font_family
        current_font_size = self._document.font_size_pt
        current_font_style = self._document.font_style
        if color == "red":
            red = 255
        elif color == "green":
            green = 255
        elif color == "blue":
            blue = 255

        self._document.set_text_color(red, green, blue)
        self._document.set_font_size(4)
        self._document.text(pos_x, pos_y + height, text)
        self._document.text_color = orig_color
        # self._document.set_xy(orig_x, orig_y)
        self._document.set_font(current_font_family, current_font_style, current_font_size)

    def draw_image(
        self, pos_x: int, pos_y: int, width: int, height: int, xml: "Element", file_name: str
    ) -> None:
        """
        Draw image onto current page.

        @param pos_x. left position
        @param pos_y. top position
        @param width. width
        @param height. heigth
        @param xml. Related XML section
        @param file_name. filename of temp data to use
        """
        import os

        if not file_name.lower().endswith(".png"):
            file_name_ = self._parser_tools.parseKey(file_name)
            if file_name_ is None:
                return
            file_name = file_name_

        if os.path.exists(file_name):
            pos_x = self.calculateLeftStart(pos_x)
            width = self.calculateWidth(width, pos_x)

            self._document.image(file_name, pos_x, pos_y, width, height, "PNG")

    def draw_barcode(
        self, pos_x: int, pos_y: int, width: int, height: int, xml: "Element", text: str
    ) -> None:
        """
        Draw barcode onto currrent page.
        """
        if text == "None":
            return
        from pineboolib.fllegacy import flcodbar

        file_name = application.PROJECT.tmpdir
        file_name += "/%s.png" % (text)
        codbartype = xml.get("CodBarType")

        if not os.path.exists(file_name):
            bar_code = flcodbar.FLCodBar(text)  # Code128
            if codbartype is not None:
                type: int = bar_code.nameToType(codbartype.lower())
                bar_code.setType(type)

            bar_code.setText(text)

            pix = bar_code.pixmap()
            if not pix.isNull():
                pix.save(file_name, "PNG")

        self.draw_image(pos_x + 10, pos_y, width - 20, height, xml, file_name)

    def setPageFormat(self, xml: "Element") -> None:
        """
        Define page parameters.

        @param xml: XML with KUT data
        """
        custom_size = None

        self._bottom_margin = int(xml.get("BottomMargin") or "0")
        self._left_margin = int(xml.get("LeftMargin") or "0")
        self._right_margin = int(xml.get("RightMargin") or "0")
        self._top_margin = int(xml.get("TopMargin") or "0")

        page_size = int(xml.get("PageSize") or "0")
        page_orientation = xml.get("PageOrientation") or "0"

        if page_size in [30, 31]:
            custom_size = [
                int(xml.get("CustomHeightMM") or "0"),
                int(xml.get("CustomWidthMM") or "0"),
            ]

        self._page_orientation = "P" if page_orientation == "0" else "L"
        self._page_size = self._parser_tools.convertPageSize(
            page_size, int(page_orientation), custom_size
        )  # devuelve un array

    def draw_margins(self) -> None:
        """Draw margins on the report."""

        top_margin = self._top_margin or 0
        bottom_margin = self._bottom_margin or 0
        left_margin = self._left_margin or 0
        right_margin = self._right_margin or 0
        width = self._page_size[0]
        height = self._page_size[1]

        tools = kparsertools.KParserTools()

        if top_margin:
            top_margin = tools.ratio_correction_h(top_margin)
        if bottom_margin:
            bottom_margin = tools.ratio_correction_h(bottom_margin)
        # if left_margin:
        #    left_margin = tools.ratio_correction_w(left_margin)
        if right_margin:
            right_margin = tools.ratio_correction_w(right_margin)

        self.draw_debug_line(left_margin, 0, left_margin, height)  # Vertical izquierda
        self.draw_debug_line(
            width - right_margin, 0, width - right_margin, height
        )  # Vertical derecha
        self.draw_debug_line(0, top_margin, width, top_margin)  # Horizontal superior
        self.draw_debug_line(
            0, height - bottom_margin, width, height - bottom_margin
        )  # Horizontal inferior

    def draw_debug_line(
        self,
        pos_x1: int,
        pos_y1: int,
        pos_x2: int,
        pos_y2: int,
        title: Optional[str] = None,
        color: str = "GREY",
    ) -> None:
        """Draw a debug line on the report."""
        dash_length = 2
        space_length = 2

        red = 0
        green = 0
        blue = 0
        if color == "GREY":
            red = 220
            green = 220
            blue = 220
        self._document.set_line_width(1)
        self._document.set_draw_color(red, green, blue)
        self._document.dashed_line(pos_x1, pos_y1, pos_x2, pos_y2, dash_length, space_length)

    def number_pages(self) -> int:
        """Get number of pages on the report."""
        return self._actual_append_page_no if self._actual_append_page_no > 0 else 0

    def reset_page_no(self) -> None:
        """Reset page number."""
        self._actual_append_page_no = -1



"""Test_pngroupbyquery module."""

import unittest
from pineboolib.application.database import pngroupbyquery


class TestPNGroupByQuery(unittest.TestCase):
    """TestPNGroupByQuery Class."""

    def test_full(self) -> None:
        """Test full"""

        group_by = pngroupbyquery.PNGroupByQuery(1, "prueba")
        self.assertEqual(group_by.level(), 1)
        self.assertEqual(group_by.field(), "prueba")



"""
Tests for loader.projectconfig.
"""
import os.path
import unittest
import tempfile
from unittest.mock import Mock, patch
from pineboolib.loader.projectconfig import ProjectConfig, VERSION_1_1, VERSION_1_2
from pineboolib.loader.projectconfig import PasswordMismatchError
from pineboolib.loader.tests import fixture_read

# from unittest.mock import patch, Mock


class TestProjectConfig(unittest.TestCase):
    """Test ProjectConfig class."""

    maxDiff = 50000

    def test_basic(self) -> None:
        """Test to create projectConfig class."""
        cfg = ProjectConfig(database="mydb", type="SQLite3 (SQLITE3)")
        self.assertTrue(cfg)
        self.assertEqual(cfg.SAVE_VERSION, VERSION_1_2)

    def test_read_write(self) -> None:
        """Test that we can read a file, save it back, read it again and stays the same."""
        project_test1 = fixture_read("project_test1.xml")
        with tempfile.TemporaryDirectory() as tmpdirname:
            cfg = ProjectConfig(
                database="mydb",
                type="SQLite3 (SQLITE3)",
                filename=os.path.join(tmpdirname, "test.xml"),
            )
            cfg.SAVE_VERSION = VERSION_1_1
            cfg.save_projectxml(False)
            file1_ = open(cfg.filename)
            data1_ = file1_.read()
            file1_.close()
            self.assertEqual(data1_, project_test1)
            cfg2 = ProjectConfig(load_xml=cfg.filename)
            cfg2.SAVE_VERSION = cfg2.version
            cfg2.save_projectxml(True)
            file2_ = open(cfg2.filename)
            data2_ = file2_.read()
            file2_.close()
            self.assertEqual(data2_, project_test1)

    @patch("time.time")
    @patch("os.urandom")
    def test_read_write2(self, mock_urandom: Mock, mock_time: Mock) -> None:
        """Test we can read and write and stays equal (slightly more complicated)."""
        # NOTE: urandom and time need to be mocked so hashes and ciphers always return the same value
        mock_urandom.side_effect = lambda n: b"1234567890123456789012345678901234567890"[:n]
        mock_time.side_effect = lambda: 10000
        project_test2 = fixture_read("project_test2.xml")
        project_test3 = fixture_read("project_test3.xml")
        with tempfile.TemporaryDirectory() as tmpdirname:
            # Verify Version 1.1
            cfg = ProjectConfig(
                database="postgres_testdb",
                description="Postgres Test DB",
                type="PostgreSQL (PSYCOPG2)",
                host="192.168.1.101",
                port=5432,
                username="postgres",
                password="postgrespassword",
                project_password="myhardtoguesspassword",
                filename=os.path.join(tmpdirname, "test.xml"),
            )
            cfg.SAVE_VERSION = VERSION_1_1
            cfg.save_projectxml(False)
            file0_ = open(cfg.filename)
            data0_ = file0_.read()
            file0_.close()
            self.assertEqual(data0_, project_test2)

            with self.assertRaises(PasswordMismatchError):
                ProjectConfig(load_xml=cfg.filename, project_password="wrongpassword")

            cfg2 = ProjectConfig(load_xml=cfg.filename, project_password="myhardtoguesspassword")
            cfg2.SAVE_VERSION = cfg2.version
            cfg2.save_projectxml(True)
            file1_ = open(cfg2.filename)
            data1_ = file1_.read()
            file1_.close()
            self.assertEqual(data1_, project_test2)

            # Verify Version 1.2
            cfg2.SAVE_VERSION = VERSION_1_2
            cfg2.save_projectxml(True)
            file2_ = open(cfg2.filename)
            data2_ = file2_.read()
            file2_.close()
            # print(open(cfg2.filename).read())
            self.assertEqual(data2_, project_test3)

            with self.assertRaises(PasswordMismatchError):
                ProjectConfig(load_xml=cfg2.filename, project_password="wrongpassword")

            cfg3 = ProjectConfig(load_xml=cfg2.filename, project_password="myhardtoguesspassword")
            cfg3.SAVE_VERSION = VERSION_1_2
            cfg3.save_projectxml(True)
            file3_ = open(cfg3.filename)
            data3_ = file3_.read()
            file3_.close()
            self.assertEqual(data3_, project_test3)



"""Utils module."""

from typing import List, Any, TYPE_CHECKING


from pineboolib.application import qsadictmodules
from pineboolib.core.utils import logging
import sqlalchemy  # type: ignore [import]

if TYPE_CHECKING:
    from sqlalchemy.ext import declarative  # type: ignore [import] # pragma: no cover
    from sqlalchemy.orm import query  # type: ignore [import] # pragma: no cover
    from sqlalchemy.orm import Session


LOGGER = logging.get_logger(__name__)


class OrmManager(object):
    """OrmManager class."""

    def __getattr__(self, name: str) -> Any:
        """Return model."""

        return qsadictmodules.QSADictModules.orm_(name)

    def models(self) -> List[str]:
        """Return available models list."""
        result_list: List[str] = []
        for name in list(dir(qsadictmodules.QSADictModules.qsa_dict_modules())):
            if str(name).endswith("_orm"):
                result_list.append(name[:-4])

        return result_list


class DynamicFilter(object):
    """DynamicFilter class."""

    query_: "query.Query"
    model_class: "declarative.DeclarativeMeta"
    filter_condition: List[List[str]]
    order_by: List[List[str]]

    def __init__(
        self,
        query: "query.Query",
        model_class: "declarative.DeclarativeMeta",
        filter_condition: List[List[str]] = [],
    ):
        """Initialize."""

        self.query_ = query
        self.model_class = model_class
        self.filter_condition = filter_condition
        self.order_by = []

    # def get_query(self):
    #    """
    #    Returns query with all the objects
    #    :return:
    #    """
    # if not self.query_:
    #    self.query_ = self.session.query(self.model_class)
    #    return self.query_

    def set_filter_condition_from_string(self, filter_str: str) -> None:
        """Set filter condition from string."""
        filter_list: List[List[str]] = []
        order_by_list = []
        order_by_str: str = ""

        pos_order = filter_str.lower().find("order by")
        if pos_order > -1:
            order_by_str = filter_str[pos_order + 9 :].lower()
            filter_str = filter_str[:pos_order]

            for order in order_by_str.split(","):
                order = order.strip()
                order_by_list.append(order.split(" "))

        if filter_str:
            filter_str = filter_str.replace("<=", " le ")
            filter_str = filter_str.replace(">=", " ge ")
            filter_str = filter_str.replace("=", " eq ")
            filter_str = filter_str.replace("<", " lt ")
            filter_str = filter_str.replace(">", " gt ")
            filter_str = filter_str.replace(" in ", " in_ ")

            item = []
            pasa = 0

            list_ = filter_str.split(" ")

            for number, part in enumerate(list_):
                if part.find("_|_space_|_") > -1:
                    part = part.replace("_|_space_|_", " ")

                if pasa:
                    pasa -= 1
                    continue

                if not part:
                    continue

                if part.startswith("upper("):
                    item.append("upper")
                    part = part[6:-1]
                elif part.find("'") > -1:
                    pos_ini = part.find("'")
                    while part[pos_ini + 1 :].find("'") == -1:
                        part = "%s %s" % (part, list_[number + 1])
                        pasa += 1

                    part = part.replace("'", "")

                elif part.lower() in ["and", "or"] and list_[number - 2] != "eq":
                    filter_list.append(item)
                    item = []
                    part = part.lower()

                item.append(part)
            if item and item != ["1", "eq", "1"]:
                filter_list.append(item)

        # =======================================================================
        # print(
        #     "\nConvirtiendo:",
        #     filter_str,
        #     "\nOrder by:",
        #     order_by_list,
        #     "\nActual:",
        #     filter_list,
        #     "\nORDER:",
        #     order_by_list,
        # )
        # =======================================================================

        self.order_by = order_by_list
        self.filter_condition = filter_list

        # ===============================================================================
        #         filter_list = []
        #         try:
        #             for key, filter in self.where_filters.items():
        #                 if not filter:
        #                     continue
        #                 # filter = filter.lower()
        #                 filter = filter.replace("=", "eq")
        #                 filter = filter.replace("<=", "le")
        #                 filter = filter.replace(">=", "ge")
        #                 filter = filter.replace("<", "lt")
        #                 filter = filter.replace(">", "gt")
        #                 filter = filter.replace(" in ", " in_ ")
        #
        #                 item = filter.split(" ")
        #                 for number, part in enumerate(item):
        #                     if part.startswith("upper("):
        #                         item[number] = part[6:-1]
        #
        #                     if part.startswith("'"):
        #                         item[number] = part[1:-1]
        #                 filter_list.append(item)
        #         except Exception as error:
        #             LOGGER.warning(
        #                 "creando filtro %s : %s", self.where_filters, str(error), stack_info=True
        #             )
        #
        #         print("**", filter_list)
        # ===============================================================================
        # print("Filtro final", self.filter_condition)

    def filter_query(
        self, query_: "query.Query", filter_condition: List[List[str]]
    ) -> "query.Query":
        """
        Return filtered queryset based on condition.

        :param query: takes query
        :param filter_condition: Its a list, ie: [(key,operator,value)]
        operator list:
            eq for ==
            lt for <
            ge for >=
            in for in_
            like for like
            value could be list or a string
        :return: queryset

        """

        # if query is None:
        #    query = self.get_query()
        # model_class = self.get_model_class()  # returns the query's Model
        model_class = self.model_class
        for raw in filter_condition:
            try:
                func_ = ""
                func_class = None
                extra_filter = ""

                if len(raw) == 3:
                    key, option, value = raw
                elif len(raw) == 4:
                    func_or_extra, key, option, value = raw

                    if func_or_extra in ["and", "or"]:
                        extra_filter = func_or_extra
                    else:
                        func_ = func_or_extra

                elif len(raw) == 5:
                    extra_filter, func_, key, option, value = raw
                else:
                    raise Exception("arguments length error", raw)  # pragma: no cover

            except ValueError:  # pragma: no cover
                raise Exception("Invalid filter: %s" % raw)  # pragma: no cover
            column = getattr(model_class, key, None)
            try:
                if func_:
                    func_class = getattr(sqlalchemy.func, func_)
            except Exception:  # pragma: no cover
                raise Exception("Error parsing func_")  # pragma: no cover

            if not column:  # pragma: no cover
                raise Exception("Invalid filter column: %s" % key, raw)  # pragma: no cover
            if option == "in":
                if isinstance(value, list):  # type: ignore [unreachable] # noqa: F821
                    filt = column.in_(value)  # type: ignore [unreachable] # noqa: F821
                else:
                    filt = column.in_(value.split(","))
            else:
                try:
                    attr = (
                        list(
                            filter(lambda e: hasattr(column, e % option), ["%s", "%s_", "__%s__"])
                        )[0]
                        % option
                    )
                except IndexError:  # pragma: no cover
                    raise Exception("Invalid filter operator: %s" % option)  # pragma: no cover
                if value == "null":
                    value = ""

                filt = getattr(column if not func_class else func_class(column), attr)(value)

            if extra_filter not in ["and", ""]:
                if extra_filter == "or":
                    filt = sqlalchemy.or_(filt)
                else:
                    raise Exception("Unknown extra filter", extra_filter)  # pragma: no cover

            query_ = query_.filter(filt)

        for name, ord in self.order_by:
            column_order = getattr(model_class, name)
            query_ = query_.order_by(column_order.desc() if ord == "desc" else column_order.asc())

        return query_

    def return_query(self) -> "query.Query":
        """Return query object."""

        return self.filter_query(self.query_, self.filter_condition)


def do_flush(session: "Session", obj_list: List[Any]) -> None:
    """Flush object on a session."""

    before_len = len(session.dirty) + len(session.new) + len(session.deleted)
    dirty_before = [obj for obj in session.dirty if obj not in obj_list]
    session_objs = [
        obj_
        for obj_ in obj_list
        if obj_ in session.dirty or obj_ in session.deleted or obj_ in session.new
    ]
    changed_len = len(session_objs)
    expected_len = before_len - changed_len

    if changed_len != len(obj_list):
        LOGGER.warning(
            "Las siguientes instancias hacen save sin estar modificadas: %s"
            % [obj for obj in obj_list if obj not in session_objs]
        )

    # Explicación. Si realizamos flush , y no hay transacción, autocommit creará una transacción y borrará todo le que existe en dirty!
    need_revert_autocommit = False
    if (
        dirty_before and session.autocommit
    ):  # Si hay objetos en dirty que no hacen flush ahora y está activado autocommit, autocommit a False.
        session.autocommit = False
        need_revert_autocommit = True

    session.flush(obj_list)

    if (
        not dirty_before
    ) and need_revert_autocommit:  # Entiendo que no hay mas registro en dirty que pueda fastidiarle el autocommit y este estaba activo
        session.autocommit = True

    after_len = len(session.dirty) + len(session.new) + len(session.deleted)

    if after_len != expected_len:
        raise Exception(
            "La cantidad de instancias restantes (%s) al realizar do_flush es diferente a la esperada (%s): "
            % (after_len, expected_len)
            + "\nInstancias pendientes de flush en session actual: nuevas: %s, modificadas: %s (previas: %s), borradas: %s"
            % (
                session.new,
                session.dirty,
                dirty_before,
                session.deleted,
            )
        )



"""Qcheckbox module."""
# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets  # type: ignore[import]


class QCheckBox(QtWidgets.QCheckBox):
    """QCheckBox class."""

    _parent: QtWidgets.QWidget

    def __init__(self, *args) -> None:
        """Inicialize."""

        if len(args) == 1:
            parent = args[0]
        else:
            self.setObjectName(args[0])
            parent = args[1]

        super().__init__(parent)

    def get_checked(self) -> bool:
        """Return if checked."""

        return self.isChecked()

    def set_checked(self, value: bool) -> None:
        """Set checked."""

        if isinstance(value, str):
            value = value == "true"

        super().setChecked(value)

    checked = property(get_checked, set_checked)



# -*- coding: utf-8 -*-
"""
AQSmtpClient Module.

This class provides utilities to use a Smtp client.
"""

from pineboolib.fllegacy.flsmtpclient import FLSmtpClient


class AQSmtpClient(FLSmtpClient):
    """AQSmtpClient Class."""

    pass



"""Flposprinter module."""

# -*- coding: utf-8 -*-
from pineboolib.core import decorators
from typing import Any, List

P76MM = 76
P57_5MM = 57
P69_5MM = 69


class FLPosPrinter(object):
    """FLPosPrinter Class."""

    _printer_name: str
    _str_buffer: List[str]
    _esc_buffer: List[str]
    _idx_buffer: List[str]
    _papers_width: List[int]
    _paper_width: int
    server_: str
    _queue_name: str

    def __init__(self) -> None:
        """Inicialize."""

        self._str_buffer = []
        self._esc_buffer = []
        self._idx_buffer = []

        self._papers_width = [P57_5MM, P69_5MM, P76MM]
        self._paper_width = P76MM

    def __del__(self) -> None:
        """Destroyer."""

        self.cleanup()

    def paperWidths(self) -> List[int]:
        """Return page widths."""

        return self._papers_width

    def paperWidth(self) -> int:
        """Return the current page width."""

        return self._paper_width

    def setPaperWidth(self, paper_width: int) -> None:
        """Set the paper width."""

        self._paper_width = paper_width

    def printerName(self) -> str:
        """Return the name of the printer."""

        return self._printer_name

    @decorators.not_implemented_warn
    def metric(self, metric: Any):
        """Not implemented."""

        pass

    def setPrinterName(self, name: str) -> None:
        """Set the name of the printer."""

        self._printer_name = name

    @decorators.beta_implementation
    def cleanup(self) -> None:
        """Clean buffer values."""

        if self._str_buffer:
            self._str_buffer = []

        if self._idx_buffer:
            self._idx_buffer = []

        self._idx_buffer = []

    @decorators.not_implemented_warn
    def flush(self):
        """Not implemented."""
        pass

    @decorators.not_implemented_warn
    def send(self, str_: str, col: int = -1, row: int = -1):
        """Not implemented."""
        pass

    @decorators.not_implemented_warn
    def sendStr(self, text_: str, col: int, row: int):
        """Not implemented."""
        pass

    @decorators.not_implemented_warn
    def sendEsc(self, esc_: str, col: int, row: int):
        """Not implemented."""
        pass

    @decorators.not_implemented_warn
    def cmd(self, command_: str, paint: Any, params: List[Any]):
        """Not implemented."""
        pass

    @decorators.beta_implementation
    def paperWidthToCols(self) -> int:
        """Return the number of columns from the paper width."""

        ret = -1
        if self._paper_width is P76MM:
            ret = 80
        elif self._paper_width is P69_5MM:
            ret = 65
        elif self._paper_width is P57_5MM:
            ret = 55
        return ret

    @decorators.not_implemented_warn
    def initFile(self):
        """Not implemented."""
        pass

    @decorators.beta_implementation
    def initStrBuffer(self) -> None:
        """Initialize the _str_buffer buffer."""

        if not self._str_buffer:
            self._str_buffer = []
        else:
            self._str_buffer.clear()

    @decorators.beta_implementation
    def initEscBuffer(self) -> None:
        """Initialize the _esc_buffer buffer."""
        if not self._esc_buffer:
            self._esc_buffer = []
        else:
            self._esc_buffer.clear()

    @decorators.beta_implementation
    def parsePrinterName(self) -> None:
        """Resolve values ​​from the printer name."""

        posdots = self._printer_name.find(":")
        self.server_ = self._printer_name[:posdots]
        self._queue_name = self._printer_name[posdots:]
        print("FLPosPrinter:parsePinterName", self.server_, self._queue_name)



"""
Lexer for Pineboo QS files.
"""
from ply.lex import TOKEN  # type: ignore


# Reserved words
reserved = [
    "BREAK",
    "CASE",
    "CONST",
    "STATIC",
    "CONTINUE",
    "DEFAULT",
    "DO",
    "ELSE",
    "FOR",
    "IF",
    "IN",
    "RETURN",
    "SWITCH",
    "WHILE",
    "CLASS",
    "VAR",
    "FUNCTION",
    "EXTENDS",
    "NEW",
    "WITH",
    "TRY",
    "CATCH",
    "THROW",
    "DELETE",
    "TYPEOF",
]
token_literals = [
    # Literals (identifier, integer constant, float constant, string constant, char const)
    "ID",
    "ICONST",
    "FCONST",
    "SCONST",
    "CCONST",  # , 'RXCONST'
]
tokens = (
    reserved
    + token_literals
    + [
        # Operators (+,-,*,/,%,|,&,~,^,<<,>>, ||, &&, !, <, <=, >, >=, ==, !=)
        "PLUS",
        "MINUS",
        "TIMES",
        "DIVIDE",
        "MOD",
        "OR",
        "AND",
        "AT",
        "XOR",
        "LSHIFT",
        "RSHIFT",
        "LOR",
        "LAND",
        "LNOT",
        "LT",
        "LE",
        "GT",
        "GE",
        "EQ",
        "NE",
        "EQQ",
        "NEQ",
        # Assignment (=, *=, /=, %=, +=, -=, <<=, >>=, &=, ^=, |=)
        "EQUALS",
        "TIMESEQUAL",
        "DIVEQUAL",
        "MODEQUAL",
        "PLUSEQUAL",
        "MINUSEQUAL",
        #    'LSHIFTEQUAL','RSHIFTEQUAL', 'ANDEQUAL', 'XOREQUAL', 'OREQUAL',
        # Increment/decrement (++,--)
        "PLUSPLUS",
        "MINUSMINUS",
        # Structure dereference (->)
        #    'ARROW',
        # Conditional operator (?)
        "CONDITIONAL1",
        #    'CONDOP',
        # Delimeters ( ) [ ] { } , . ; :
        "LPAREN",
        "RPAREN",
        "LBRACKET",
        "RBRACKET",
        "LBRACE",
        "RBRACE",
        "COMMA",
        "PERIOD",
        "SEMI",
        "COLON",
        # Ellipsis (...)
        #    'ELLIPSIS',
        "DOCSTRINGOPEN",
        #   'COMMENTOPEN',
        "COMMENTCLOSE",
        "DOLLAR",
        "SQOUTE",
        "DQOUTE",
        "BACKSLASH",
    ]
)

# Completely ignored characters
t_ignore = " \r\t\x0c"  # pylint: disable=invalid-name

# Newlines


@TOKEN(r"\n+")
def t_NEWLINE(t):  # pylint: disable=invalid-name
    """Keep track of line numbers."""
    t.lexer.lineno += t.value.count("\n")


# Operators
t_BACKSLASH = "\\\\"  # pylint: disable=invalid-name
t_DOLLAR = r"\$"  # pylint: disable=invalid-name
t_SQOUTE = "'"  # pylint: disable=invalid-name
t_DQOUTE = '"'  # pylint: disable=invalid-name
t_PLUS = r"\+"  # pylint: disable=invalid-name
t_MINUS = r"-"  # pylint: disable=invalid-name
t_TIMES = r"\*"  # pylint: disable=invalid-name
t_DIVIDE = r"/"  # pylint: disable=invalid-name
t_MOD = r"%"  # pylint: disable=invalid-name
t_OR = r"\|"  # pylint: disable=invalid-name
t_AND = r"&"  # pylint: disable=invalid-name
# t_NOT              = r'~'
t_XOR = r"\^"  # pylint: disable=invalid-name
t_LSHIFT = r"<<"  # pylint: disable=invalid-name
t_RSHIFT = r">>"  # pylint: disable=invalid-name
t_LOR = r"\|\|"  # pylint: disable=invalid-name
t_LAND = r"&&"  # pylint: disable=invalid-name
t_LNOT = r"!"  # pylint: disable=invalid-name
t_LT = r"<"  # pylint: disable=invalid-name
t_GT = r">"  # pylint: disable=invalid-name
t_LE = r"<="  # pylint: disable=invalid-name
t_GE = r">="  # pylint: disable=invalid-name
t_EQ = r"=="  # pylint: disable=invalid-name
t_NE = r"!="  # pylint: disable=invalid-name
t_EQQ = r"==="  # pylint: disable=invalid-name
t_NEQ = r"!=="  # pylint: disable=invalid-name
t_CONDITIONAL1 = r"\?"  # pylint: disable=invalid-name

# Assignment operators

t_EQUALS = r"="  # pylint: disable=invalid-name
t_TIMESEQUAL = r"\*="  # pylint: disable=invalid-name
t_DIVEQUAL = r"/="  # pylint: disable=invalid-name
t_MODEQUAL = r"%="  # pylint: disable=invalid-name
t_PLUSEQUAL = r"\+="  # pylint: disable=invalid-name
t_MINUSEQUAL = r"-="  # pylint: disable=invalid-name

# Increment/decrement
t_PLUSPLUS = r"\+\+"  # pylint: disable=invalid-name
t_MINUSMINUS = r"--"  # pylint: disable=invalid-name

# ->
# t_ARROW            = r'->'

# ?
# t_CONDOP           = r'\?'


# Delimeters
t_LPAREN = r"\("  # pylint: disable=invalid-name
t_RPAREN = r"\)"  # pylint: disable=invalid-name
t_LBRACKET = r"\["  # pylint: disable=invalid-name
t_RBRACKET = r"\]"  # pylint: disable=invalid-name
t_LBRACE = r"\{"  # pylint: disable=invalid-name
t_RBRACE = r"\}"  # pylint: disable=invalid-name
t_COMMA = r","  # pylint: disable=invalid-name
t_PERIOD = r"\."  # pylint: disable=invalid-name
t_SEMI = r";"  # pylint: disable=invalid-name
t_COLON = r":"  # pylint: disable=invalid-name
# t_ELLIPSIS         = r'\.\.\.'
t_AT = r"@"  # pylint: disable=invalid-name
# Identifiers and reserved words

reserved_map = {}
for r in reserved:
    reserved_map[r.lower()] = r


@TOKEN(r"[A-Za-z_]+[\w_]*")
def t_ID(t):  # pylint: disable=invalid-name
    """Get ID tokens."""
    t.type = reserved_map.get(t.value, "ID")
    return t


# Integer literal
t_ICONST = r"\d+([uU]|[lL]|[uU][lL]|[lL][uU])?"  # pylint: disable=invalid-name

# Floating literal
t_FCONST = (  # pylint: disable=invalid-name
    r"((\d+)(\.\d+)(e(\+|-)?(\d+))? | (\d+)e(\+|-)?(\d+))([lL]|[fF])?"
)

# String literal
t_SCONST = r"\"([^\"\\\n]|(\\.)|\\\n)*?\""  # pylint: disable=invalid-name

# Character constant 'c' or L'c'
t_CCONST = r"\'([^\'\\\n]|(\\.)|\\\n)*?\'"  # pylint: disable=invalid-name

# REGEX constant
# t_RXCONST = r'/[^/ ]+/g?'

# Comments


@TOKEN(r"(/\*( |\*\*)(.|\n)*?\*/)|(//.*)")
def t_comment(t):  # pylint: disable=invalid-name
    """Keep track of line count in comments."""
    t.lexer.lineno += t.value.count("\n")


@TOKEN(r"/\*\*[ ]+")
def t_DOCSTRINGOPEN(t):  # pylint: disable=invalid-name
    """Return docstring for later analysis."""
    return t


# t_COMMENTOPEN      = r'/\*'
t_COMMENTCLOSE = r"\*/"  # pylint: disable=invalid-name


# Preprocessor directive (ignored)
@TOKEN(r"\#(.)*?\n")
def t_preprocessor(t):  # pylint: disable=invalid-name
    """Ignored Preprocessor directive. Not used."""
    t.lexer.lineno += 1


def t_error(t) -> None:  # pylint: disable=invalid-name
    """Skip invalid characters and report."""
    print("Illegal character %s" % repr(t.value[0]))
    t.lexer.skip(1)



"""Splash_screen module."""
from PyQt6 import QtGui, QtCore, QtWidgets  # type: ignore
from pineboolib.core.utils.utils_base import filedir
from pineboolib.core import settings


class SplashScreen(object):
    """Show a splashscreen to inform keep the user busy while Pineboo is warming up."""

    _splash: "QtWidgets.QSplashScreen"

    def __init__(self):
        """Inicialize."""
        splash_path = filedir(
            "./core/images/splashscreen/%s240.png"
            % ("dbadmin" if settings.CONFIG.value("application/dbadmin_enabled") else "quick")
        )

        splash_pix = QtGui.QPixmap(splash_path)
        self._splash = QtWidgets.QSplashScreen(
            splash_pix, QtCore.Qt.WindowType.WindowStaysOnTopHint
        )
        self._splash.setMask(splash_pix.mask())

        frame_geo = self._splash.frameGeometry()
        primary_screen = QtGui.QGuiApplication.primaryScreen()
        if primary_screen:
            frame_geo.moveCenter(primary_screen.geometry().center())
        self._splash.move(frame_geo.topLeft())

    def showMessage(self, text: str) -> None:
        """Show a message into spalsh screen."""
        self._splash.showMessage(
            text, QtCore.Qt.AlignmentFlag.AlignLeft, QtCore.Qt.GlobalColor.white
        )

    def hide(self) -> None:
        """Hide splash screen."""
        QtCore.QTimer.singleShot(1000, self._splash.hide)

    def show(self) -> None:
        """Show splash screen."""
        self._splash.show()



# -*- coding: utf-8 -*-
# Translated with pineboolib v0.71.18
"""Flvar_model module."""

import sqlalchemy  # type: ignore [import] # noqa: F821

from pineboolib.application.database.orm import basemodel


class Flvar(basemodel.BaseModel):  # type: ignore [misc] # noqa: F821
    """Flvar class."""

    __tablename__ = "flvar"

    # --- Metadata --->
    legacy_metadata = {
        "name": "flvar",
        "alias": "Variables",
        "fields": [
            {
                "name": "id",
                "alias": "Identificador",
                "pk": True,
                "type": "serial",
                "null": False,
                "visiblegrid": False,
            },
            {
                "name": "idvar",
                "alias": "Identificador de la variable",
                "ck": True,
                "type": "string",
                "length": 30,
                "null": False,
            },
            {
                "name": "idsesion",
                "alias": "Identificador de la sesión",
                "ck": True,
                "type": "string",
                "length": 30,
                "null": False,
            },
            {
                "name": "valor",
                "alias": "Valor",
                "type": "stringlist",
                "null": False,
                "visiblegrid": False,
            },
        ],
    }

    # <--- Metadata ---

    # --- Fields --->

    id = sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True)
    idvar = sqlalchemy.Column("idvar", sqlalchemy.String(30))
    idsesion = sqlalchemy.Column("idsesion", sqlalchemy.String(30))
    valor = sqlalchemy.Column("valor", sqlalchemy.String)


# <--- Fields ---



"""
test_pnfieldmetadata Module.
"""

import unittest
from pineboolib.loader.main import init_testing

from pineboolib import application


class TestStringField(unittest.TestCase):
    """Test string field."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test collect a field from the pntablemetadata flmodules and check the values"""

        mtd = application.PROJECT.conn_manager.manager().metadata("flmodules")
        self.assertTrue(mtd is not None)
        if mtd is not None:
            field = mtd.field("version")
            self.assertNotEqual(field, None)
            if field is not None:
                self.assertEqual(field.name(), "version")
                self.assertEqual(field.isPrimaryKey(), False)
                self.assertEqual(field.isCompoundKey(), False)
                self.assertEqual(field.length(), 3)
                self.assertEqual(field.allowNull(), False)
                self.assertEqual(field.visibleGrid(), True)
                self.assertEqual(field.visible(), True)
                self.assertEqual(field.editable(), False)
                self.assertEqual(field.defaultValue(), "0.0")
                self.assertEqual(field.regExpValidator(), r"[0-9]\\.[0-9]")

                assign_value_1 = field.formatAssignValue("version", "a.1", False)
                assign_value_2 = field.formatAssignValue("version", "b.1", True)
                self.assertEqual(assign_value_1, "version = 'a.1'")
                self.assertEqual(assign_value_2, "upper(version) = 'B.1'")


class TestCopyField(unittest.TestCase):
    """Test Copy a field."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test copy a field data from another."""

        from pineboolib.application.metadata import pnfieldmetadata

        mtd = application.PROJECT.conn_manager.manager().metadata("flmodules")
        if mtd is not None:
            field_1 = mtd.field("version")

            field_2 = pnfieldmetadata.PNFieldMetaData(field_1)

            self.assertNotEqual(field_2, None)
            self.assertEqual(field_2.name(), "version")
            self.assertEqual(field_2.isPrimaryKey(), False)
            self.assertEqual(field_2.isCompoundKey(), False)
            self.assertEqual(field_2.length(), 3)
            self.assertEqual(field_2.allowNull(), False)
            self.assertEqual(field_2.visibleGrid(), True)
            self.assertEqual(field_2.visible(), True)
            self.assertEqual(field_2.editable(), False)
            self.assertEqual(field_2.defaultValue(), "0.0")
            self.assertEqual(field_2.regExpValidator(), r"[0-9]\\.[0-9]")


class TestUintField(unittest.TestCase):
    """Test uint field."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test collect a field from the pntablemetadata flseq and check the values"""

        mtd = application.PROJECT.conn_manager.manager().metadata("flseqs")
        self.assertTrue(mtd is not None)
        if mtd is not None:
            field = mtd.field("seq")
            self.assertTrue(field is not None)
            if field is not None:
                self.assertNotEqual(field, None)
                self.assertEqual(field.name(), "seq")
                self.assertEqual(field.alias(), "Secuencia")
                self.assertEqual(field.isPrimaryKey(), False)
                self.assertEqual(field.isCompoundKey(), False)
                self.assertEqual(field.length(), 0)
                self.assertEqual(field.allowNull(), False)
                self.assertEqual(field.visibleGrid(), True)
                self.assertEqual(field.visible(), True)
                self.assertEqual(field.editable(), True)
                self.assertEqual(field.defaultValue(), None)
                self.assertEqual(field.regExpValidator(), None)

                assign_value = field.formatAssignValue("seq", 666, False)

                self.assertEqual(assign_value, "seq = 666")

                table_metadata = field.metadata()
                self.assertTrue(table_metadata is not None)
                if table_metadata is not None:
                    self.assertEqual(table_metadata.name(), mtd.name())


class TestStringListField(unittest.TestCase):
    """Test stringlist field."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test collect a field from the pntablemetadata flsettings and check the values"""

        mtd = application.PROJECT.conn_manager.manager().metadata("flsettings")
        self.assertTrue(mtd)
        if mtd is not None:
            field = mtd.field("valor")
            self.assertTrue(field is not None)
            if field is not None:
                self.assertNotEqual(field, None)
                self.assertEqual(field.name(), "valor")
                self.assertEqual(field.alias(), "Valor")
                self.assertEqual(field.isPrimaryKey(), False)
                self.assertEqual(field.isCompoundKey(), False)
                self.assertEqual(field.length(), 0)
                self.assertEqual(field.allowNull(), True)
                self.assertEqual(field.visibleGrid(), False)
                self.assertEqual(field.visible(), True)
                self.assertEqual(field.editable(), True)
                self.assertEqual(field.defaultValue(), None)
                self.assertEqual(field.regExpValidator(), None)
                self.assertEqual(field.flDecodeType("stringlist"), "string")


class TestPixmapField(unittest.TestCase):
    """Test pixmap field."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test collect a field from the pntablemetadata flmodules and check the values"""

        mtd = application.PROJECT.conn_manager.manager().metadata("flmodules")
        self.assertTrue(mtd)
        if mtd is not None:
            field = mtd.field("icono")
            if field is not None:
                self.assertNotEqual(field, None)
                self.assertEqual(field.name(), "icono")
                self.assertEqual(field.isPrimaryKey(), False)
                self.assertEqual(field.isCompoundKey(), False)
                self.assertEqual(field.length(), 0)
                self.assertEqual(field.allowNull(), True)
                self.assertEqual(field.type(), "pixmap")
                self.assertEqual(field.visibleGrid(), True)
                self.assertEqual(field.visible(), True)
                self.assertEqual(field.editable(), True)
                self.assertEqual(field.defaultValue(), None)
                self.assertEqual(field.flDecodeType("pixmap"), "string")


class TestUnlockField(unittest.TestCase):
    """Test unlock field."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test collect a field from the pntablemetadata flmodules and check the values"""

        mtd = application.PROJECT.conn_manager.manager().metadata("flmodules")
        self.assertTrue(mtd)
        if mtd is not None:
            field = mtd.field("bloqueo")
            if field is not None:
                self.assertNotEqual(field, None)
                self.assertEqual(field.name(), "bloqueo")
                self.assertEqual(field.isPrimaryKey(), False)
                self.assertEqual(field.isCompoundKey(), False)
                self.assertEqual(field.length(), 0)
                self.assertEqual(field.allowNull(), False)
                self.assertEqual(field.type(), "unlock")
                self.assertEqual(field.visibleGrid(), True)
                self.assertEqual(field.visible(), True)
                self.assertEqual(field.editable(), True)
                self.assertEqual(field.defaultValue(), True)
                self.assertEqual(field.regExpValidator(), None)
                self.assertEqual(field.flDecodeType("unlock"), "bool")


class TestBoolField(unittest.TestCase):
    """Test bool field."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test collect a field from the pntablemetadata flmetadata and check the values"""

        mtd = application.PROJECT.conn_manager.manager().metadata("flmetadata")
        self.assertTrue(mtd)
        if mtd is not None:
            field = mtd.field("bloqueo")
            self.assertNotEqual(field, None)
            if field is not None:
                self.assertEqual(field.name(), "bloqueo")
                self.assertEqual(field.alias(), "Tabla bloqueada")
                self.assertEqual(field.isPrimaryKey(), False)
                self.assertEqual(field.isCompoundKey(), False)
                self.assertEqual(field.length(), 0)
                self.assertEqual(field.allowNull(), True)
                self.assertEqual(field.type(), "bool")
                self.assertEqual(field.visibleGrid(), True)
                self.assertEqual(field.visible(), True)
                self.assertEqual(field.editable(), True)
                self.assertEqual(field.defaultValue(), None)
                self.assertEqual(field.regExpValidator(), None)
                self.assertEqual(field.flDecodeType("bool"), "bool")


class TestDateField(unittest.TestCase):
    """Test date field."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test collect a field from the pntablemetadata flupdates and check the values"""

        mtd = application.PROJECT.conn_manager.manager().metadata("flupdates")
        self.assertTrue(mtd)
        if mtd is not None:
            field = mtd.field("fecha")
            self.assertNotEqual(field, None)
            if field is not None:
                self.assertEqual(field.name(), "fecha")
                self.assertEqual(field.alias(), "Fecha")
                self.assertEqual(field.isPrimaryKey(), False)
                self.assertEqual(field.isCompoundKey(), False)
                self.assertEqual(field.length(), 0)
                self.assertEqual(field.type(), "date")
                self.assertEqual(field.allowNull(), False)
                self.assertEqual(field.visibleGrid(), True)
                self.assertEqual(field.visible(), True)
                self.assertEqual(field.editable(), False)
                self.assertEqual(field.defaultValue(), None)
                self.assertEqual(field.regExpValidator(), None)
                self.assertEqual(field.flDecodeType("date"), "date")


class TestTimeField(unittest.TestCase):
    """Test time field."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test collect a field from the pntablemetadata flupdates and check the values"""

        mtd = application.PROJECT.conn_manager.manager().metadata("flupdates")
        self.assertTrue(mtd)
        if mtd is not None:
            field = mtd.field("hora")
            self.assertNotEqual(field, None)
            if field is not None:
                self.assertEqual(field.name(), "hora")
                self.assertEqual(field.alias(), "Hora")
                self.assertEqual(field.isPrimaryKey(), False)
                self.assertEqual(field.isCompoundKey(), False)
                self.assertEqual(field.length(), 0)
                self.assertEqual(field.type(), "time")
                self.assertEqual(field.allowNull(), False)
                self.assertEqual(field.visibleGrid(), True)
                self.assertEqual(field.visible(), True)
                self.assertEqual(field.editable(), False)
                self.assertEqual(field.defaultValue(), None)
                self.assertEqual(field.regExpValidator(), None)
                self.assertEqual(field.flDecodeType("time"), "time")


class TestDoubleField(unittest.TestCase):
    """Test double field."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test collect a field and check the values"""

        from pineboolib.application.metadata import pnfieldmetadata

        field = pnfieldmetadata.PNFieldMetaData(
            "new_double",
            "Nuevo Double",
            False,
            False,
            "double",
            0,
            False,
            True,
            True,
            5,
            8,
            False,
            False,
            False,
            0.01,
            False,
            None,
            True,
            False,
            False,
        )

        self.assertNotEqual(field, None)
        self.assertEqual(field.name(), "new_double")
        self.assertEqual(field.alias(), "Nuevo Double")
        self.assertEqual(field.isPrimaryKey(), False)
        self.assertEqual(field.isCompoundKey(), False)
        self.assertEqual(field.length(), 0)
        self.assertEqual(field.type(), "double")
        self.assertEqual(field.allowNull(), False)
        self.assertEqual(field.visibleGrid(), True)
        self.assertEqual(field.visible(), True)
        self.assertEqual(field.editable(), True)
        self.assertEqual(field.defaultValue(), 0.01)
        self.assertEqual(field.regExpValidator(), None)
        self.assertEqual(field.partInteger(), 5)
        self.assertEqual(field.partDecimal(), 8)
        self.assertEqual(field.flDecodeType("double"), "double")


class TestOptionsListField(unittest.TestCase):
    """Test string field with optionsList."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test collect a field and check the values"""

        from pineboolib.application.metadata import pnfieldmetadata

        field = pnfieldmetadata.PNFieldMetaData(
            "new_string",
            "Nuevo String",
            False,
            False,
            "string",
            20,
            False,
            True,
            True,
            0,
            0,
            False,
            False,
            False,
            "primero",
            False,
            None,
            True,
            False,
            False,
        )

        self.assertNotEqual(field, None)

        field.setOptionsList("primero,segundo,tercero,cuarto,quinto,sexto,12345678901234567890")

        self.assertEqual(field.name(), "new_string")
        self.assertEqual(field.alias(), "Nuevo String")
        self.assertEqual(field.isPrimaryKey(), False)
        self.assertEqual(field.isCompoundKey(), False)
        self.assertEqual(field.length(), 20)
        self.assertEqual(field.type(), "string")
        self.assertEqual(field.allowNull(), False)
        self.assertEqual(field.visibleGrid(), True)
        self.assertEqual(field.visible(), True)
        self.assertEqual(field.editable(), True)
        self.assertEqual(field.defaultValue(), "primero")
        self.assertEqual(field.regExpValidator(), None)
        self.assertEqual(field.partInteger(), 0)
        self.assertEqual(field.partDecimal(), 0)
        self.assertEqual(field.flDecodeType("string"), "string")
        self.assertEqual(
            field.optionsList(),
            ["primero", "segundo", "tercero", "cuarto", "quinto", "sexto", "12345678901234567890"],
        )
        self.assertEqual(field.getIndexOptionsList("segundo"), 1)
